# future
from __future__ import annotations

# stdlib
import operator
import typing
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple as TypeTuple
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey
import numpy as np
import numpy.typing as npt
import torch

# relative
from .... import lib
from ....ast.klass import pointerize_args_and_kwargs
from ....proto.core.adp.phi_tensor_pb2 import (
    SingleEntityPhiTensor as SingleEntityPhiTensor_PB,
)
from ....proto.core.tensor.single_entity_phi_tensor_pb2 import (
    TensorWrappedSingleEntityPhiTensorPointer as TensorWrappedSingleEntityPhiTensorPointer_PB,
)
from ....util import inherit_tags
from ...adp.entity import Entity
from ...adp.vm_private_scalar_manager import VirtualMachinePrivateScalarManager
from ...common.serde.deserialize import _deserialize as deserialize
from ...common.serde.serializable import serializable
from ...common.serde.serialize import _serialize as serialize
from ...common.uid import UID
from ...node.abstract.node import AbstractNodeClient
from ...node.common.action.run_class_method_action import RunClassMethodAction
from ...pointer.pointer import Pointer
from ..ancestors import AutogradTensorAncestor
from ..broadcastable import is_broadcastable
from ..passthrough import AcceptableSimpleType  # type: ignore
from ..passthrough import PassthroughTensor  # type: ignore
from ..passthrough import SupportedChainType  # type: ignore
from ..passthrough import implements  # type: ignore
from ..passthrough import is_acceptable_simple_type  # type: ignore
from ..smpc import utils
from ..smpc.mpc_tensor import MPCTensor
from ..smpc.share_tensor import ShareTensor
from ..smpc.utils import TYPE_TO_RING_SIZE
from ..tensor import Tensor
from ..util import inputs2child  # type: ignore
from .adp_tensor import ADPTensor
from .dp_tensor_converter import convert_to_gamma_tensor
from .initial_gamma import InitialGammaTensor
from .initial_gamma import IntermediateGammaTensor


@serializable()
class TensorWrappedSingleEntityPhiTensorPointer(Pointer):
    """
    This tensor represents a pointer to a very specific tensor chain. Eventually we'll have some sort
    of more intelligent/general representation for pointers to chains of objects, but for now this is
    what we're going with. This pointer represents all the arguments of the objects in the chain as its
    attributes.

    Thus, this class has two groups of attributes: one set are the attributes for SingeEntityPhiTensor:
        child: SupportedChainType,
        entity: Entity,
        min_vals: np.ndarray,
        max_vals: np.ndarray,
        scalar_manager: Optional[VirtualMachinePrivateScalarManager] = None,

    And the others are for initializing a Pointer object:
        client=self.client,
        id_at_location=self.id_at_location,
        object_type=self.object_type,
        tags=self.tags,
        description=self.description,
    """

    __name__ = "TensorWrappedSingleEntityPhiTensorPointer"
    __module__ = "syft.core.tensor.autodp.single_entity_phi"

    def __init__(
        self,
        entity: Entity,
        min_vals: np.typing.ArrayLike,
        max_vals: np.typing.ArrayLike,
        client: Any,
        scalar_manager: Optional[VirtualMachinePrivateScalarManager] = None,
        id_at_location: Optional[UID] = None,
        object_type: str = "",
        tags: Optional[List[str]] = None,
        description: str = "",
        public_shape: Optional[TypeTuple[int, ...]] = None,
        public_dtype: Optional[np.dtype] = None,
    ):

        super().__init__(
            client=client,
            id_at_location=id_at_location,
            object_type=object_type,
            tags=tags,
            description=description,
        )

        self.min_vals = min_vals
        self.max_vals = max_vals
        self.entity = entity
        self.scalar_manager = scalar_manager
        self.public_shape = public_shape
        self.public_dtype = public_dtype

    @property
    def synthetic(self) -> np.ndarray:
        return (
            np.random.rand(*list(self.public_shape)) * (self.max_vals - self.min_vals)  # type: ignore
            + self.min_vals
        ).astype(self.public_dtype)

    def __repr__(self) -> str:
        return (
            self.synthetic.__repr__()
            + "\n\n (The data printed above is synthetic - it's an imitation of the real data.)"
        )

    def share(self, *parties: TypeTuple[AbstractNodeClient, ...]) -> MPCTensor:
        all_parties = list(parties) + [self.client]
        ring_size = TYPE_TO_RING_SIZE.get(self.public_dtype, None)
        self_mpc = MPCTensor(
            secret=self,
            shape=self.public_shape,
            ring_size=ring_size,
            parties=all_parties,
        )
        return self_mpc

    def _apply_tensor_op(self, other: Any, op_str: str) -> Any:
        # we want to get the return type which matches the attr_path_and_name
        # so we ask lib_ast for the return type name that matches out
        # attr_path_and_name and then use that to get the actual pointer klass
        # then set the result to that pointer klass

        attr_path_and_name = f"syft.core.tensor.tensor.Tensor.__{op_str}__"

        result = TensorWrappedSingleEntityPhiTensorPointer(
            entity=self.entity,
            min_vals=self.min_vals,
            max_vals=self.max_vals,
            client=self.client,
            scalar_manager=self.scalar_manager,
        )

        # QUESTION can the id_at_location be None?
        result_id_at_location = getattr(result, "id_at_location", None)

        if result_id_at_location is not None:
            # first downcast anything primitive which is not already PyPrimitive
            (
                downcast_args,
                downcast_kwargs,
            ) = lib.python.util.downcast_args_and_kwargs(args=[other], kwargs={})

            # then we convert anything which isnt a pointer into a pointer
            pointer_args, pointer_kwargs = pointerize_args_and_kwargs(
                args=downcast_args,
                kwargs=downcast_kwargs,
                client=self.client,
                gc_enabled=False,
            )

            cmd = RunClassMethodAction(
                path=attr_path_and_name,
                _self=self,
                args=pointer_args,
                kwargs=pointer_kwargs,
                id_at_location=result_id_at_location,
                address=self.client.address,
            )
            self.client.send_immediate_msg_without_reply(msg=cmd)

        inherit_tags(
            attr_path_and_name=attr_path_and_name,
            result=result,
            self_obj=self,
            args=[other],
            kwargs={},
        )

        result_public_shape = None

        if isinstance(other, TensorWrappedSingleEntityPhiTensorPointer):
            other_shape = other.public_shape
            other_dtype = other.public_dtype
        elif isinstance(other, (int, float)):
            other_shape = (1,)
            other_dtype = np.int32
        elif isinstance(other, bool):
            other_shape = (1,)
            other_dtype = np.dtype("bool")
        elif isinstance(other, np.ndarray):
            other_shape = other.shape
            other_dtype = other.dtype
        else:
            raise ValueError(
                f"Invalid Type for TensorWrappedSingleEntityPhiTensorPointer:{type(other)}"
            )

        if self.public_shape is not None and other_shape is not None:
            result_public_shape = utils.get_shape(
                op_str, self.public_shape, other_shape
            )

        if self.public_dtype is not None and other_dtype is not None:
            if self.public_dtype != other_dtype:
                raise ValueError(
                    f"Type for self and other do not match ({self.public_dtype} vs {other_dtype})"
                )
            result_public_dtype = self.public_dtype

        result.public_shape = result_public_shape
        result.public_dtype = result_public_dtype

        return result

    @staticmethod
    def _apply_op(
        self: TensorWrappedSingleEntityPhiTensorPointer,
        other: Union[
            TensorWrappedSingleEntityPhiTensorPointer, MPCTensor, int, float, np.ndarray
        ],
        op_str: str,
    ) -> Union[MPCTensor, TensorWrappedSingleEntityPhiTensorPointer]:
        """Performs the operation based on op_str

        Args:
            other (Union[TensorWrappedSingleEntityPhiTensorPointer,MPCTensor,int,float,np.ndarray]): second operand.

        Returns:
            Tuple[MPCTensor,Union[MPCTensor,int,float,np.ndarray]] : Result of the operation
        """
        op = getattr(operator, op_str)

        if (
            isinstance(other, TensorWrappedSingleEntityPhiTensorPointer)
            and self.client != other.client
        ):

            parties = [self.client, other.client]

            self_mpc = MPCTensor(secret=self, shape=self.public_shape, parties=parties)
            other_mpc = MPCTensor(
                secret=other, shape=other.public_shape, parties=parties
            )

            return op(self_mpc, other_mpc)

        elif isinstance(other, MPCTensor):

            return op(other, self)

        return self._apply_tensor_op(other=other, op_str=op_str)

    def __add__(
        self,
        other: Union[
            TensorWrappedSingleEntityPhiTensorPointer, MPCTensor, int, float, np.ndarray
        ],
    ) -> Union[TensorWrappedSingleEntityPhiTensorPointer, MPCTensor]:
        """Apply the "add" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedSingleEntityPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedSingleEntityPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorWrappedSingleEntityPhiTensorPointer._apply_op(self, other, "add")

    def __sub__(
        self,
        other: Union[
            TensorWrappedSingleEntityPhiTensorPointer, MPCTensor, int, float, np.ndarray
        ],
    ) -> Union[TensorWrappedSingleEntityPhiTensorPointer, MPCTensor]:
        """Apply the "sub" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedSingleEntityPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedSingleEntityPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorWrappedSingleEntityPhiTensorPointer._apply_op(self, other, "sub")

    def __mul__(
        self,
        other: Union[
            TensorWrappedSingleEntityPhiTensorPointer, MPCTensor, int, float, np.ndarray
        ],
    ) -> Union[TensorWrappedSingleEntityPhiTensorPointer, MPCTensor]:
        """Apply the "mul" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedSingleEntityPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedSingleEntityPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorWrappedSingleEntityPhiTensorPointer._apply_op(self, other, "mul")

    def __lt__(
        self,
        other: Union[
            TensorWrappedSingleEntityPhiTensorPointer, MPCTensor, int, float, np.ndarray
        ],
    ) -> Union[TensorWrappedSingleEntityPhiTensorPointer, MPCTensor]:
        """Apply the "lt" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedSingleEntityPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedSingleEntityPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorWrappedSingleEntityPhiTensorPointer._apply_op(self, other, "lt")

    def __gt__(
        self,
        other: Union[
            TensorWrappedSingleEntityPhiTensorPointer, MPCTensor, int, float, np.ndarray
        ],
    ) -> Union[TensorWrappedSingleEntityPhiTensorPointer, MPCTensor]:
        """Apply the "gt" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedSingleEntityPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedSingleEntityPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorWrappedSingleEntityPhiTensorPointer._apply_op(self, other, "gt")

    def __ge__(
        self,
        other: Union[
            TensorWrappedSingleEntityPhiTensorPointer, MPCTensor, int, float, np.ndarray
        ],
    ) -> Union[TensorWrappedSingleEntityPhiTensorPointer, MPCTensor]:
        """Apply the "ge" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedSingleEntityPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedSingleEntityPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorWrappedSingleEntityPhiTensorPointer._apply_op(self, other, "ge")

    def __le__(
        self,
        other: Union[
            TensorWrappedSingleEntityPhiTensorPointer, MPCTensor, int, float, np.ndarray
        ],
    ) -> Union[TensorWrappedSingleEntityPhiTensorPointer, MPCTensor]:
        """Apply the "le" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedSingleEntityPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedSingleEntityPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorWrappedSingleEntityPhiTensorPointer._apply_op(self, other, "le")

    def __eq__(  # type: ignore
        self,
        other: Union[
            TensorWrappedSingleEntityPhiTensorPointer, MPCTensor, int, float, np.ndarray
        ],
    ) -> Union[TensorWrappedSingleEntityPhiTensorPointer, MPCTensor]:
        """Apply the "eq" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedSingleEntityPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedSingleEntityPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorWrappedSingleEntityPhiTensorPointer._apply_op(self, other, "eq")

    def __ne__(  # type: ignore
        self,
        other: Union[
            TensorWrappedSingleEntityPhiTensorPointer, MPCTensor, int, float, np.ndarray
        ],
    ) -> Union[TensorWrappedSingleEntityPhiTensorPointer, MPCTensor]:
        """Apply the "ne" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedSingleEntityPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedSingleEntityPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorWrappedSingleEntityPhiTensorPointer._apply_op(self, other, "ne")

    def to_local_object_without_private_data_child(self) -> SingleEntityPhiTensor:
        """Convert this pointer into a partial version of the SingleEntityPhiTensor but without
        any of the private data therein."""

        public_shape = getattr(self, "public_shape", None)
        public_dtype = getattr(self, "public_dtype", None)
        return Tensor(
            child=SingleEntityPhiTensor(
                child=None,
                entity=self.entity,
                min_vals=self.min_vals,  # type: ignore
                max_vals=self.max_vals,  # type: ignore
                scalar_manager=self.scalar_manager,
            ),
            public_shape=public_shape,
            public_dtype=public_dtype,
        )

    def _object2proto(self) -> "TensorWrappedSingleEntityPhiTensorPointer_PB":

        _entity = serialize(self.entity)
        _min_vals = serialize(self.min_vals)
        _max_vals = serialize(self.max_vals)
        _location = serialize(self.client.address)
        _scalar_manager = serialize(self.scalar_manager, to_bytes=True)
        _id_at_location = serialize(self.id_at_location)
        _object_type = self.object_type
        _tags = self.tags
        _description = self.description
        _public_shape = serialize(getattr(self, "public_shape", None), to_bytes=True)
        _public_dtype = serialize(getattr(self, "public_dtype", None), to_bytes=True)

        return TensorWrappedSingleEntityPhiTensorPointer_PB(
            entity=_entity,
            min_vals=_min_vals,
            max_vals=_max_vals,
            location=_location,
            scalar_manager=_scalar_manager,
            id_at_location=_id_at_location,
            object_type=_object_type,
            tags=_tags,
            description=_description,
            public_shape=_public_shape,
            public_dtype=_public_dtype,
        )

    @staticmethod
    def _proto2object(
        proto: TensorWrappedSingleEntityPhiTensorPointer_PB,
    ) -> "TensorWrappedSingleEntityPhiTensorPointer":

        entity = deserialize(blob=proto.entity)
        min_vals = deserialize(blob=proto.min_vals)
        max_vals = deserialize(blob=proto.max_vals)
        client = deserialize(blob=proto.location)
        scalar_manager = deserialize(blob=proto.scalar_manager, from_bytes=True)
        id_at_location = deserialize(blob=proto.id_at_location)
        object_type = proto.object_type
        tags = proto.tags
        public_shape = deserialize(blob=proto.public_shape, from_bytes=True)
        public_dtype = deserialize(blob=proto.public_dtype, from_bytes=True)
        description = proto.description

        return TensorWrappedSingleEntityPhiTensorPointer(
            entity=entity,
            min_vals=min_vals,
            max_vals=max_vals,
            client=client,
            scalar_manager=scalar_manager,
            id_at_location=id_at_location,
            object_type=object_type,
            tags=tags,
            description=description,
            public_shape=public_shape,
            public_dtype=public_dtype,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return TensorWrappedSingleEntityPhiTensorPointer_PB


@serializable()
class SingleEntityPhiTensor(PassthroughTensor, AutogradTensorAncestor, ADPTensor):

    PointerClassOverride = TensorWrappedSingleEntityPhiTensorPointer

    # Number of entities in a SEPT is by definition 1
    n_entities = 1
    __slots__ = (
        "child",
        "_min_vals",
        "_max_vals",
        "entity",
        "scalar_manager",
    )

    def __init__(
        self,
        child: SupportedChainType,
        entity: Entity,
        min_vals: np.ndarray,
        max_vals: np.ndarray,
        scalar_manager: Optional[VirtualMachinePrivateScalarManager] = None,
    ) -> None:

        # child = the actual private data
        super().__init__(child)

        # identically shaped tensor to "child" but making the LOWEST possible value of this private value
        self._min_vals = min_vals

        # identically shaped tensor to "child" but making the HIGHEST possible value of this private value
        self._max_vals = max_vals

        # the identity of the data subject
        self.entity = entity

        if scalar_manager is None:
            self.scalar_manager = VirtualMachinePrivateScalarManager()
        else:
            self.scalar_manager = scalar_manager

    def init_pointer(
        self,
        client: Any,
        id_at_location: Optional[UID] = None,
        object_type: str = "",
        tags: Optional[List[str]] = None,
        description: str = "",
    ) -> TensorWrappedSingleEntityPhiTensorPointer:
        return TensorWrappedSingleEntityPhiTensorPointer(
            # Arguments specifically for SEPhiTensor
            entity=self.entity,
            min_vals=self._min_vals,
            max_vals=self._max_vals,
            scalar_manager=self.scalar_manager,
            # Arguments required for a Pointer to work
            client=client,
            id_at_location=id_at_location,
            object_type=object_type,
            tags=tags,
            description=description,
        )

    @property
    def gamma(self) -> InitialGammaTensor:
        """Property to cast this tensor into a GammaTensor"""
        return self.create_gamma()

    def create_gamma(
        self, scalar_manager: Optional[VirtualMachinePrivateScalarManager] = None
    ) -> InitialGammaTensor:
        """Return a new Gamma tensor based on this phi tensor"""

        if scalar_manager is None:
            scalar_manager = self.scalar_manager

        # Gamma expects an entity for each scalar
        entities = np.array([self.entity] * np.array(self.child.shape).prod()).reshape(
            self.shape
        )

        return InitialGammaTensor(
            values=self.child,
            min_vals=self.min_vals,
            max_vals=self.max_vals,
            entities=entities,
            scalar_manager=scalar_manager,
        )

    def publish(
        self, acc: Any, sigma: float, user_key: VerifyKey
    ) -> AcceptableSimpleType:
        print("PUBLISHING TO GAMMA:")
        print(self.child)
        return self.gamma.publish(acc=acc, sigma=sigma, user_key=user_key)

    @property
    def min_vals(self) -> np.ndarray:
        return self._min_vals

    @property
    def max_vals(self) -> np.ndarray:
        return self._max_vals

    def astype(self, np_type: np.dtype) -> SingleEntityPhiTensor:
        return self.__class__(
            child=self.child.astype(np_type),
            entity=self.entity,
            min_vals=self.min_vals.astype(np_type),
            max_vals=self.max_vals.astype(np_type),
            scalar_manager=self.scalar_manager,
        )

    def __repr__(self) -> str:

        """Pretty print some information, optimized for Jupyter notebook viewing."""
        return (
            f"{self.__class__.__name__}(entity={self.entity.name}, child={self.child})"
        )

    def __and__(self, other: Any) -> SingleEntityPhiTensor:
        """Note: this is bitwise and, not logical and"""
        return SingleEntityPhiTensor(
            child=self.child & other,
            min_vals=np.zeros_like(self.child),
            max_vals=np.ones_like(self.child),
            entity=self.entity,
        )

    def __or__(self, other: Any) -> SingleEntityPhiTensor:
        return SingleEntityPhiTensor(
            child=self.child | other,
            min_vals=np.zeros_like(self.child),
            max_vals=np.ones_like(self.child),
            entity=self.entity,
        )

    # Check for shape1 = (1,s), and shape2 = (,s) --> as an example
    def __eq__(
        self, other: SupportedChainType
    ) -> Union[SingleEntityPhiTensor, IntermediateGammaTensor]:
        if is_acceptable_simple_type(other):
            if isinstance(other, np.ndarray):
                # If other is a Numpy Array, we need to check if shapes can be broadcast
                if is_broadcastable(other.shape, self.child.shape):  # type: ignore
                    data = self.child == other
                else:
                    raise Exception(
                        f"Tensor shapes do not match "  # type: ignore
                        f"for __eq__: {self.child.shape} != {other.child.shape}"  # type: ignore
                    )
            else:
                data = self.child == other
        elif isinstance(other, SingleEntityPhiTensor):
            if self.entity != other.entity:
                return convert_to_gamma_tensor(self) == convert_to_gamma_tensor(other)
            else:
                if is_broadcastable(self.child.shape, other.child.shape):  # type: ignore
                    data = self.child == other.child
                else:
                    raise Exception(
                        f"Tensor shapes do not match for __eq__: {self.child.shape} != {other.child.shape}"
                    )
        elif isinstance(other, PassthroughTensor):
            if is_broadcastable(self.child.shape, other.child.shape):  # type: ignore
                data = self.child == other.child
        else:
            raise Exception(
                f"Tensor shapes do not match for __eq__: {self.child} != len{other}"
            )
        return SingleEntityPhiTensor(
            child=data,
            entity=self.entity,
            min_vals=self.min_vals * 0.0,
            max_vals=self.max_vals * 0.0 + 1.0,
            scalar_manager=self.scalar_manager,
        )

    def __ne__(
        self, other: SupportedChainType
    ) -> Union[SingleEntityPhiTensor, IntermediateGammaTensor]:
        # Make use of the equal operator we just implemented, and invert the result
        opposite_result = self.__eq__(other)

        if isinstance(opposite_result, SingleEntityPhiTensor):
            return SingleEntityPhiTensor(
                child=opposite_result.child ^ 1,
                entity=opposite_result.entity,
                min_vals=opposite_result.min_vals,
                max_vals=opposite_result.max_vals,
                scalar_manager=opposite_result.scalar_manager,
            )
        elif isinstance(opposite_result, InitialGammaTensor):
            return InitialGammaTensor(
                values=opposite_result.values ^ 1,
                entities=opposite_result.entities,
                min_vals=opposite_result.min_vals,
                max_vals=opposite_result.max_vals,
                scalar_manager=opposite_result.scalar_manager,
            )
        else:
            raise Exception

    def __abs__(self) -> SingleEntityPhiTensor:

        data = self.child.abs()

        # create true/false gate inputs
        minvals_is_gt0 = self.min_vals > 0
        minvals_is_le0 = -minvals_is_gt0 + 1
        maxvals_is_gt0 = self.max_vals >= 0
        maxvals_is_le0 = -maxvals_is_gt0 + 1

        # create true/false gates
        is_strict_gt0 = minvals_is_gt0
        is_gtlt0 = minvals_is_le0 * maxvals_is_gt0
        is_strict_lt0 = minvals_is_le0 * maxvals_is_le0

        # if min_vals > 0, then new min_vals doesn't change
        min_vals_strict_gt0 = self.min_vals

        # if min_vals < 0 and max_vals > 0, then new min_vals = 0
        min_vals_gtlt0 = self.min_vals * 0

        # if min_vals < 0 and max_vals < 0, then new min_vals = -max_vals
        min_vals_strict_lt0 = -self.max_vals

        # sum of masked options
        min_vals = is_strict_gt0 * min_vals_strict_gt0
        min_vals = min_vals + (is_gtlt0 * min_vals_gtlt0)
        min_vals = min_vals + (is_strict_lt0 * min_vals_strict_lt0)

        #  if min_vals > 0, then new min_vals doesn't change
        max_vals_strict_gt0 = self.max_vals

        # if min_vals < 0 and max_vals > 0, then new min_vals = 0
        max_vals_gtlt0 = np.max([self.max_vals, -self.min_vals])  # type: ignore

        #  if min_vals < 0 and max_vals < 0, then new min_vals = -max_vals
        max_vals_strict_lt0 = -self.min_vals

        # sum of masked options
        max_vals = is_strict_gt0 * max_vals_strict_gt0
        max_vals = max_vals + (is_gtlt0 * max_vals_gtlt0)
        max_vals = max_vals + (is_strict_lt0 * max_vals_strict_lt0)

        entity = self.entity

        return SingleEntityPhiTensor(
            child=data,
            entity=entity,
            min_vals=min_vals,
            max_vals=max_vals,
            scalar_manager=self.scalar_manager,
        )

    def __add__(
        self, other: SupportedChainType
    ) -> Union[SingleEntityPhiTensor, IntermediateGammaTensor]:

        # if the tensor being added is also private
        if isinstance(other, SingleEntityPhiTensor):
            if self.entity != other.entity:
                left = convert_to_gamma_tensor(self)
                right = convert_to_gamma_tensor(other)
                return left + right

            data = self.child + other.child
            min_vals = self.min_vals + other.min_vals
            max_vals = self.max_vals + other.max_vals
            entity = self.entity

            return SingleEntityPhiTensor(
                child=data,
                entity=entity,
                min_vals=min_vals,
                max_vals=max_vals,
                scalar_manager=self.scalar_manager,
            )

        # if the tensor being added is a public tensor / int / float / etc.
        elif is_acceptable_simple_type(other):

            data = self.child + other
            min_vals = self.min_vals + other
            max_vals = self.max_vals + other
            entity = self.entity

            return SingleEntityPhiTensor(
                child=data,
                entity=entity,
                min_vals=min_vals,
                max_vals=max_vals,
                scalar_manager=self.scalar_manager,
            )
        elif isinstance(other, IntermediateGammaTensor):
            return convert_to_gamma_tensor(self) + other
        else:
            print("Type is unsupported:" + str(type(other)))
            raise NotImplementedError

    def __neg__(self) -> SingleEntityPhiTensor:

        data = self.child * -1
        min_vals = self.max_vals * -1
        max_vals = self.min_vals * -1
        entity = self.entity

        return SingleEntityPhiTensor(
            child=data,
            entity=entity,
            min_vals=min_vals,
            max_vals=max_vals,
            scalar_manager=self.scalar_manager,
        )

    def __getitem__(self, key: Any) -> SingleEntityPhiTensor:

        data = self.child.__getitem__(key)
        min_vals = self.min_vals.__getitem__(key)
        max_vals = self.max_vals.__getitem__(key)

        if isinstance(data, (np.number, bool, int, float)):
            data = np.array([data])  # 1 dimensional np.array
        if isinstance(min_vals, (np.number, bool, int, float)):
            min_vals = np.array(min_vals)  # 1 dimensional np.array
        if isinstance(max_vals, (np.number, bool, int, float)):
            max_vals = np.array(max_vals)  # 1 dimensional np.array

        entity = self.entity

        return SingleEntityPhiTensor(
            child=data,
            entity=entity,
            min_vals=min_vals,
            max_vals=max_vals,
            scalar_manager=self.scalar_manager,
        )

    def __mul__(
        self, other: SupportedChainType
    ) -> Union[SingleEntityPhiTensor, IntermediateGammaTensor]:

        if isinstance(other, SingleEntityPhiTensor):
            print(f"SELF ENTITY:{self.entity.name}")
            print(f"OTHER ENTITY:{other.entity.name}")
            if self.entity != other.entity:
                print("Entities are not the same?!?!?!")
                return convert_to_gamma_tensor(self) * convert_to_gamma_tensor(other)

            data = self.child * other.child

            min_min = self.min_vals * other.min_vals
            min_max = self.min_vals * other.max_vals
            max_min = self.max_vals * other.min_vals
            max_max = self.max_vals * other.max_vals

            min_vals = np.min([min_min, min_max, max_min, max_max], axis=0)  # type: ignore
            max_vals = np.max([min_min, min_max, max_min, max_max], axis=0)  # type: ignore
            entity = self.entity

            return SingleEntityPhiTensor(
                child=data,
                entity=entity,
                min_vals=min_vals,
                max_vals=max_vals,
                scalar_manager=self.scalar_manager,
            )
        elif is_acceptable_simple_type(other):

            data = self.child * other

            min_min = self.min_vals * other
            min_max = self.min_vals * other
            max_min = self.max_vals * other
            max_max = self.max_vals * other

            min_vals = np.min([min_min, min_max, max_min, max_max], axis=0)  # type: ignore
            max_vals = np.max([min_min, min_max, max_min, max_max], axis=0)  # type: ignore
            entity = self.entity

            return SingleEntityPhiTensor(
                child=data,
                entity=entity,
                min_vals=min_vals,
                max_vals=max_vals,
                scalar_manager=self.scalar_manager,
            )

        else:
            return NotImplemented

    def __pos__(self) -> SingleEntityPhiTensor:
        """Identity operator, returns itself"""
        return self

    def __sub__(
        self, other: SupportedChainType
    ) -> Union[SingleEntityPhiTensor, IntermediateGammaTensor]:

        if isinstance(other, SingleEntityPhiTensor):
            if self.entity != other.entity:
                return convert_to_gamma_tensor(self) - convert_to_gamma_tensor(other)

            data = self.child - other.child
            min_min = self.min_vals - other.min_vals
            min_max = self.min_vals - other.max_vals
            max_min = self.max_vals - other.min_vals
            max_max = self.max_vals - other.max_vals
            min_vals = np.minimum.reduce([min_min, min_max, max_min, max_max])
            max_vals = np.maximum.reduce([min_min, min_max, max_min, max_max])

            entity = self.entity

        elif is_acceptable_simple_type(other):
            if isinstance(other, np.ndarray):
                if not is_broadcastable(other.shape, self.child.shape):  # type: ignore
                    raise Exception(
                        f"Shapes do not match for subtraction: {self.child.shape} and {other.shape}"
                    )
            data = self.child - other
            min_vals = self.min_vals - other
            max_vals = self.max_vals - other
            entity = self.entity
        else:
            raise NotImplementedError
        return SingleEntityPhiTensor(
            child=data,
            entity=entity,
            min_vals=min_vals,
            max_vals=max_vals,
            scalar_manager=self.scalar_manager,
        )

    def __truediv__(self, other: SupportedChainType) -> SingleEntityPhiTensor:

        if isinstance(other, SingleEntityPhiTensor):

            if self.entity != other.entity:
                # this should return a GammaTensor
                return NotImplemented

            data = self.child / other.child

            if (other.min_vals == 0).any() or (other.max_vals == 0).any():

                raise Exception(
                    "Infinite sensitivity - we can support this in the future but not yet"
                )

            else:

                min_min = self.min_vals / other.min_vals
                min_max = self.min_vals / other.max_vals
                max_min = self.max_vals / other.min_vals
                max_max = self.max_vals / other.max_vals

                min_vals = np.min([min_min, min_max, max_min, max_max], axis=0)  # type: ignore
                max_vals = np.max([min_min, min_max, max_min, max_max], axis=0)  # type: ignore

            entity = self.entity

            return SingleEntityPhiTensor(
                child=data,
                entity=entity,
                min_vals=min_vals,
                max_vals=max_vals,
                scalar_manager=self.scalar_manager,
            )
        else:
            # Ignoring unsupported operand error b/c other logic is taken care of
            return self * (1 / other)  # type: ignore

    def compress(
        self,
        condition: np.typing.ArrayLike,
        axis: Optional[int] = None,
        out: Optional[np.ndarray] = None,
    ) -> SingleEntityPhiTensor:
        """Return selected slices of this array along a given axis"""
        if (
            isinstance(self.child, int)
            or isinstance(self.child, float)
            or isinstance(self.child, bool)
        ):
            # For these data types, the compress operation is meaningless, so don't change them.
            data = self.child
            print(
                f"Warning: Tensor data was of type {type(data)}, compress operation had no effect."
            )
        else:
            data = self.child.compress(condition, axis, out)

            # TODO: Should we give warnings for min_val and max_val being single floats/integers/booleans too?
        if (
            isinstance(self.min_vals, int)
            or isinstance(self.min_vals, float)
            or isinstance(self.min_vals, bool)
        ):
            # For these data types, the compress operation is meaningless, so don't change them.
            min_vals = self.min_vals
            # print(f'Warning: Tensor data was of type {type(data)}, compress operation had no effect.')
        else:
            min_vals = self.min_vals.compress(condition, axis, out)

        if (
            isinstance(self.max_vals, int)
            or isinstance(self.max_vals, float)
            or isinstance(self.max_vals, bool)
        ):
            # For these data types, the compress operation is meaningless, so don't change them.
            max_vals = self.max_vals
            # print(f'Warning: Tensor data was of type {type(data)}, compress operation had no effect.')
        else:
            max_vals = self.max_vals.compress(condition, axis, out)

        entity = self.entity

        return SingleEntityPhiTensor(
            child=data,
            entity=entity,
            min_vals=min_vals,
            max_vals=max_vals,
            scalar_manager=self.scalar_manager,
        )

    def dot(self, other: SupportedChainType) -> SingleEntityPhiTensor:
        return self.manual_dot(other)

    # ndarray.flatten(order='C')
    def flatten(self, order: str = "C") -> SingleEntityPhiTensor:
        if (
            isinstance(self.child, int)
            or isinstance(self.child, float)
            or isinstance(self.child, bool)
        ):
            # For these data types, the flatten operation is meaningless, so don't change them.
            data = self.child
            print(
                f"Warning: Tensor data was of type {type(data)}, flatten operation had no effect."
            )
        else:
            data = self.child.flatten(order)

            # TODO: Should we give warnings for min_val and max_val being single floats/integers/booleans too?
        if (
            isinstance(self.min_vals, int)
            or isinstance(self.min_vals, float)
            or isinstance(self.min_vals, bool)
        ):
            # For these data types, the flatten operation is meaningless, so don't change them.
            min_vals = self.min_vals
            # print(f'Warning: Tensor data was of type {type(data)}, flatten operation had no effect.')
        else:
            min_vals = self.min_vals.flatten(order)

        if (
            isinstance(self.max_vals, int)
            or isinstance(self.max_vals, float)
            or isinstance(self.max_vals, bool)
        ):
            # For these data types, the flatten operation is meaningless, so don't change them.
            max_vals = self.max_vals
            # print(f'Warning: Tensor data was of type {type(data)}, flatten operation had no effect.')
        else:
            max_vals = self.max_vals.flatten(order)

        entity = self.entity

        return SingleEntityPhiTensor(
            child=data,
            entity=entity,
            min_vals=min_vals,
            max_vals=max_vals,
            scalar_manager=self.scalar_manager,
        )

    # def partition(
    #     self,
    #     kth: Union[int, List[int], np.ndarray],
    #     axis: Optional[int] = -1,
    #     kind: Optional[str] = "introselect",
    #     order: Optional[Union[str, List[str]]] = None,
    # ) -> SingleEntityPhiTensor:
    #     # this method mutates self
    #     """Interchange two axes of the Tensor"""
    #     if (
    #         isinstance(self.child, int)
    #         or isinstance(self.child, float)
    #         or isinstance(self.child, bool)
    #     ):
    #         # For these singleton data types, the partition operation is meaningless, so don't change them.
    #         print(
    #             f"Warning: Tensor data was of type {type(self.child)}, partition operation had no effect."
    #         )
    #     else:
    #         self.child.partition(kth, axis, kind, order)
    #
    #         # TODO: Should we give warnings for min_val and max_val being single floats/integers/booleans too?
    #     if (
    #         isinstance(self.min_vals, int)
    #         or isinstance(self.min_vals, float)
    #         or isinstance(self.min_vals, bool)
    #     ):
    #         # For these singleton data types, the partition operation is meaningless, so don't change them.
    #         print(
    #             f"Warning: Min_vals metadata was of type {type(self.min_vals)}, partition operation had no effect."
    #         )
    #     else:
    #         self.min_vals.partition(kth, axis, kind, order)
    #
    #     if (
    #         isinstance(self.max_vals, int)
    #         or isinstance(self.max_vals, float)
    #         or isinstance(self.max_vals, bool)
    #     ):
    #         # For these singleton data types, the partition operation is meaningless, so don't change them.
    #         print(
    #             f"Warning: Max_vals metadata was of type {type(self.max_vals)}, partition operation had no effect."
    #         )
    #     else:
    #         self.max_vals.partition(kth, axis, kind, order)
    #
    #     return self

    # ndarray.ravel(order='C')
    def ravel(self, order: str = "C") -> SingleEntityPhiTensor:
        if (
            isinstance(self.child, int)
            or isinstance(self.child, float)
            or isinstance(self.child, bool)
        ):
            # For these data types, the ravel operation is meaningless, so don't change them.
            data = self.child
            print(
                f"Warning: Tensor data was of type {type(data)}, ravel operation had no effect."
            )
        else:
            data = self.child.ravel(order)

            # TODO: Should we give warnings for min_val and max_val being single floats/integers/booleans too?
        if (
            isinstance(self.min_vals, int)
            or isinstance(self.min_vals, float)
            or isinstance(self.min_vals, bool)
        ):
            # For these data types, the ravel operation is meaningless, so don't change them.
            min_vals = self.min_vals
            # print(f'Warning: Tensor data was of type {type(data)}, ravel operation had no effect.')
        else:
            min_vals = self.min_vals.ravel(order)

        if (
            isinstance(self.max_vals, int)
            or isinstance(self.max_vals, float)
            or isinstance(self.max_vals, bool)
        ):
            # For these data types, the ravel operation is meaningless, so don't change them.
            max_vals = self.max_vals
            # print(f'Warning: Tensor data was of type {type(data)}, ravel operation had no effect.')
        else:
            max_vals = self.max_vals.ravel(order)
        entity = self.entity

        return SingleEntityPhiTensor(
            child=data,
            entity=entity,
            min_vals=min_vals,
            max_vals=max_vals,
            scalar_manager=self.scalar_manager,
        )

    def repeat(
        self, repeats: Union[int, TypeTuple[int, ...]], axis: Optional[int] = None
    ) -> SingleEntityPhiTensor:

        data = self.child.repeat(repeats, axis=axis)
        min_vals = self.min_vals.repeat(repeats, axis=axis)
        max_vals = self.max_vals.repeat(repeats, axis=axis)
        entity = self.entity

        return SingleEntityPhiTensor(
            child=data,
            entity=entity,
            min_vals=min_vals,
            max_vals=max_vals,
            scalar_manager=self.scalar_manager,
        )

    def reshape(self, *args: Any) -> SingleEntityPhiTensor:
        if (
            isinstance(self.child, int)
            or isinstance(self.child, float)
            or isinstance(self.child, bool)
        ):
            # For these data types, the reshape operation is meaningless, so don't change them.
            data = self.child
            print(
                f"Warning: Tensor data was of type {type(data)}, reshape operation had no effect."
            )
        else:
            data = self.child.reshape(*args)

            # TODO: Should we give warnings for min_val and max_val being single floats/integers/booleans too?
        if (
            isinstance(self.min_vals, int)
            or isinstance(self.min_vals, float)
            or isinstance(self.min_vals, bool)
        ):
            # For these data types, the reshape operation is meaningless, so don't change them.
            min_vals = self.min_vals
            # print(f'Warning: Tensor data was of type {type(data)}, reshape operation had no effect.')
        else:
            min_vals = self.min_vals.reshape(*args)

        if (
            isinstance(self.max_vals, int)
            or isinstance(self.max_vals, float)
            or isinstance(self.max_vals, bool)
        ):
            # For these data types, the reshape operation is meaningless, so don't change them.
            max_vals = self.max_vals
            # print(f'Warning: Tensor data was of type {type(data)}, reshape operation had no effect.')
        else:
            max_vals = self.max_vals.reshape(*args)

        entity = self.entity

        return SingleEntityPhiTensor(
            child=data,
            entity=entity,
            min_vals=min_vals,
            max_vals=max_vals,
            scalar_manager=self.scalar_manager,
        )

    def resize(
        self,
        new_shape: Union[TypeTuple[int], int, typing.Iterable],
        refcheck: bool = True,
    ) -> None:
        """Change shape and size of array, in-place."""
        if (
            isinstance(self.child, int)
            or isinstance(self.child, float)
            or isinstance(self.child, bool)
        ):
            # For these data types, the resize operation is meaningless, so don't change them.
            pass
            print(
                f"Warning: Tensor data was of type {type(self.child)}, resize operation had no effect."
            )
        else:
            # self.child = self.child.resize(new_shape, refcheck)
            self.child = np.resize(self.child, new_shape)

        # TODO: Should we give warnings for min_val and max_val being single floats/integers/booleans too?
        if (
            isinstance(self.min_vals, int)
            or isinstance(self.min_vals, float)
            or isinstance(self.min_vals, bool)
        ):
            # For these data types, the resize operation is meaningless, so don't change them.
            pass
            # print(f'Warning: min_vals data was of type {type(self.min_vals)}, resize operation had no effect.')
        else:
            self._min_vals = np.reshape(self.min_vals, new_shape)

        if (
            isinstance(self.max_vals, int)
            or isinstance(self.max_vals, float)
            or isinstance(self.max_vals, bool)
        ):
            # For these data types, the resize operation is meaningless, so don't change them.
            pass
            # print(f'Warning: max_vals data was of type {type(data)}, resize operation had no effect.')
        else:
            self._max_vals = np.reshape(self.max_vals, new_shape)

        return None

    def squeeze(self, axis: Optional[int] = None) -> SingleEntityPhiTensor:
        if (
            isinstance(self.child, int)
            or isinstance(self.child, float)
            or isinstance(self.child, bool)
        ):
            # For these data types, the squeeze operation is meaningless, so don't change them.
            data = self.child
            print(
                f"Warning: Tensor data was of type {type(data)}, squeeze operation had no effect."
            )
        else:
            data = self.child.squeeze(axis)

            # TODO: Should we give warnings for min_val and max_val being single floats/integers/booleans too?
        if (
            isinstance(self.min_vals, int)
            or isinstance(self.min_vals, float)
            or isinstance(self.min_vals, bool)
        ):
            # For these data types, the squeeze operation is meaningless, so don't change them.
            min_vals = self.min_vals
            # print(f'Warning: Tensor data was of type {type(data)}, squeeze operation had no effect.')
        else:
            min_vals = self.min_vals.squeeze(axis)

        if (
            isinstance(self.max_vals, int)
            or isinstance(self.max_vals, float)
            or isinstance(self.max_vals, bool)
        ):
            # For these data types, the squeeze operation is meaningless, so don't change them.
            max_vals = self.max_vals
            # print(f'Warning: Tensor data was of type {type(data)}, squeeze operation had no effect.')
        else:
            max_vals = self.max_vals.squeeze(axis)

        entity = self.entity

        return SingleEntityPhiTensor(
            child=data,
            entity=entity,
            min_vals=min_vals,
            max_vals=max_vals,
            scalar_manager=self.scalar_manager,
        )

    def sum(self, *args: Any, **kwargs: Any) -> SingleEntityPhiTensor:

        data = self.child.sum(*args, **kwargs)
        min_vals = self.min_vals.sum(*args, **kwargs)
        max_vals = self.max_vals.sum(*args, **kwargs)
        entity = self.entity

        data = np.array(data)
        min_vals = np.array(min_vals)
        max_vals = np.array(max_vals)

        return SingleEntityPhiTensor(
            child=data,
            entity=entity,
            min_vals=min_vals,
            max_vals=max_vals,
            scalar_manager=self.scalar_manager,
        )

    def swapaxes(self, axis1: int, axis2: int) -> SingleEntityPhiTensor:
        """Interchange two axes of the Tensor"""
        if (
            isinstance(self.child, int)
            or isinstance(self.child, float)
            or isinstance(self.child, bool)
        ):
            # For these singleton data types, the swapaxes operation is meaningless, so don't change them.
            data = self.child
            print(
                f"Warning: Tensor data was of type {type(data)}, swapaxes operation had no effect."
            )
        else:
            data = self.child.swapaxes(axis1, axis2)

            # TODO: Should we give warnings for min_val and max_val being single floats/integers/booleans too?
        if (
            isinstance(self.min_vals, int)
            or isinstance(self.min_vals, float)
            or isinstance(self.min_vals, bool)
        ):
            # For these singleton data types, the swapaxes operation is meaningless, so don't change them.
            min_vals = self.min_vals
            # print(f'Warning: Tensor data was of type {type(data)}, swapaxes operation had no effect.')
        else:
            min_vals = self.min_vals.swapaxes(axis1, axis2)

        if (
            isinstance(self.max_vals, int)
            or isinstance(self.max_vals, float)
            or isinstance(self.max_vals, bool)
        ):
            # For these singleton data types, the swapaxes operation is meaningless, so don't change them.
            max_vals = self.max_vals
            # print(f'Warning: Tensor data was of type {type(data)}, swapaxes operation had no effect.')
        else:
            max_vals = self.max_vals.swapaxes(axis1, axis2)

        entity = self.entity

        return SingleEntityPhiTensor(
            child=data,
            entity=entity,
            min_vals=min_vals,
            max_vals=max_vals,
            scalar_manager=self.scalar_manager,
        )

    def transpose(self, *args: Any, **kwargs: Any) -> SingleEntityPhiTensor:
        """Transposes self.child, min_vals, and max_vals if these can be transposed, otherwise doesn't change them."""
        if (
            isinstance(self.child, int)
            or isinstance(self.child, float)
            or isinstance(self.child, bool)
        ):
            # For these data types, the transpose operation is meaningless, so don't change them.
            data = self.child
            print(
                f"Warning: Tensor data was of type {type(data)}, transpose operation had no effect."
            )
        else:
            data = self.child.transpose(*args)

        # TODO: Should we give warnings for min_val and max_val being single floats/integers/booleans too?
        if (
            isinstance(self.min_vals, int)
            or isinstance(self.min_vals, float)
            or isinstance(self.min_vals, bool)
        ):
            # For these data types, the transpose operation is meaningless, so don't change them.
            min_vals = self.min_vals
            # print(f'Warning: Tensor data was of type {type(data)}, transpose operation had no effect.')
        else:
            min_vals = self.min_vals.transpose(*args)

        if (
            isinstance(self.max_vals, int)
            or isinstance(self.max_vals, float)
            or isinstance(self.max_vals, bool)
        ):
            # For these data types, the transpose operation is meaningless, so don't change them.
            max_vals = self.max_vals
            # print(f'Warning: Tensor data was of type {type(data)}, transpose operation had no effect.')
        else:
            max_vals = self.max_vals.transpose(*args)

        entity = self.entity

        return SingleEntityPhiTensor(
            child=data,
            entity=entity,
            min_vals=min_vals,
            max_vals=max_vals,
            scalar_manager=self.scalar_manager,
        )

    def __le__(
        self, other: SupportedChainType
    ) -> Union[SingleEntityPhiTensor, IntermediateGammaTensor]:

        # if the tensor being compared is also private
        if isinstance(other, SingleEntityPhiTensor):

            if self.entity != other.entity:
                # this should return a GammaTensor
                return convert_to_gamma_tensor(self) <= convert_to_gamma_tensor(other)

            if len(self.child) != len(other.child):
                raise Exception(
                    f"Tensor dims do not match for __le__: {len(self.child)} != {len(other.child)}"  # type: ignore
                )

            data = (
                self.child <= other.child
            ) * 1  # the * 1 just makes sure it returns integers instead of True/False
            min_vals = self.min_vals * 0
            max_vals = (self.max_vals * 0) + 1
            entity = self.entity

            return SingleEntityPhiTensor(
                child=data,
                entity=entity,
                min_vals=min_vals,
                max_vals=max_vals,
                scalar_manager=self.scalar_manager,
            )

        # if the tensor being compared is a public tensor / int / float / etc.
        elif is_acceptable_simple_type(other):

            data = (self.child <= other) * 1
            min_vals = self.min_vals * 0
            max_vals = (self.max_vals * 0) + 1
            entity = self.entity

            return SingleEntityPhiTensor(
                child=data,
                entity=entity,
                min_vals=min_vals,
                max_vals=max_vals,
                scalar_manager=self.scalar_manager,
            )

        else:
            return NotImplemented

    def __ge__(
        self, other: SupportedChainType
    ) -> Union[SingleEntityPhiTensor, IntermediateGammaTensor]:

        # if the tensor being compared is also private
        if isinstance(other, SingleEntityPhiTensor):

            if self.entity != other.entity:
                # this should return a GammaTensor
                return convert_to_gamma_tensor(self) >= convert_to_gamma_tensor(other)

            if len(self.child) != len(other.child):
                raise Exception(
                    f"Tensor dims do not match for __ge__: {len(self.child)} != {len(other.child)}"  # type: ignore
                )

            data = (
                self.child >= other.child
            ) * 1  # the * 1 just makes sure it returns integers instead of True/False
            min_vals = self.min_vals * 0
            max_vals = (self.max_vals * 0) + 1
            entity = self.entity

            return SingleEntityPhiTensor(
                child=data,
                entity=entity,
                min_vals=min_vals,
                max_vals=max_vals,
                scalar_manager=self.scalar_manager,
            )

        # if the tensor being compared is a public tensor / int / float / etc.
        elif is_acceptable_simple_type(other):

            data = (self.child >= other) * 1
            min_vals = self.min_vals * 0
            max_vals = (self.max_vals * 0) + 1
            entity = self.entity

            return SingleEntityPhiTensor(
                child=data,
                entity=entity,
                min_vals=min_vals,
                max_vals=max_vals,
                scalar_manager=self.scalar_manager,
            )

        else:
            return NotImplemented

    def __lt__(
        self, other: SupportedChainType
    ) -> Union[SingleEntityPhiTensor, IntermediateGammaTensor]:

        # if the tensor being compared is also private
        if isinstance(other, SingleEntityPhiTensor):

            if self.entity != other.entity:
                # this should return a GammaTensor
                return convert_to_gamma_tensor(self) < convert_to_gamma_tensor(other)

            if len(self.child) != len(other.child):
                raise Exception(
                    f"Tensor dims do not match for __lt__: {len(self.child)} != {len(other.child)}"  # type: ignore
                )

            data = (
                self.child < other.child
            ) * 1  # the * 1 just makes sure it returns integers instead of True/False
            min_vals = self.min_vals * 0
            max_vals = (self.max_vals * 0) + 1
            entity = self.entity

            return SingleEntityPhiTensor(
                child=data,
                entity=entity,
                min_vals=min_vals,
                max_vals=max_vals,
                scalar_manager=self.scalar_manager,
            )

        # if the tensor being compared is a public tensor / int / float / etc.
        elif is_acceptable_simple_type(other):

            data = (self.child < other) * 1
            min_vals = self.min_vals * 0
            max_vals = (self.max_vals * 0) + 1
            entity = self.entity

            return SingleEntityPhiTensor(
                child=data,
                entity=entity,
                min_vals=min_vals,
                max_vals=max_vals,
                scalar_manager=self.scalar_manager,
            )

        else:
            return NotImplemented

    def __gt__(self, other: SupportedChainType) -> SingleEntityPhiTensor:

        # if the tensor being compared is also private
        if isinstance(other, SingleEntityPhiTensor):

            if self.entity != other.entity:
                # this should return a GammaTensor
                return convert_to_gamma_tensor(self) > convert_to_gamma_tensor(other)

            if len(self.child) != len(other.child):
                raise Exception(
                    f"Tensor dims do not match for __gt__: {len(self.child)} != {len(other.child)}"  # type: ignore
                )

            data = (
                self.child > other.child
            ) * 1  # the * 1 just makes sure it returns integers instead of True/False
            min_vals = self.min_vals * 0
            max_vals = (self.max_vals * 0) + 1
            entity = self.entity

            return SingleEntityPhiTensor(
                child=data,
                entity=entity,
                min_vals=min_vals,
                max_vals=max_vals,
                scalar_manager=self.scalar_manager,
            )

        # if the tensor being compared is a public tensor / int / float / etc.
        elif is_acceptable_simple_type(other):

            data = (self.child > other) * 1
            min_vals = self.min_vals * 0
            max_vals = (self.max_vals * 0) + 1
            entity = self.entity

            return SingleEntityPhiTensor(
                child=data,
                entity=entity,
                min_vals=min_vals,
                max_vals=max_vals,
                scalar_manager=self.scalar_manager,
            )

        else:
            return NotImplemented

    def clip(
        self, a_min: npt.ArrayLike, a_max: npt.ArrayLike, *args: Any
    ) -> SingleEntityPhiTensor:

        if a_min is None and a_max is None:
            raise Exception("ValueError: clip: must set either max or min")

        if is_acceptable_simple_type(self.child):
            if isinstance(self.child, np.ndarray):
                data = self.child.clip(a_min, a_max)
            else:
                # self.child is a singleton
                data = max(a_min, min(self.child, a_max)) if a_min <= a_max else a_max
        elif isinstance(self.child, torch.Tensor):
            data = self.child.numpy().clip(a_min, a_max)

        if isinstance(self.min_vals, np.ndarray):
            min_vals = np.clip(self.min_vals, a_min=a_min, a_max=a_max, *args)
        else:
            min_vals = (
                max(a_min, min(self.min_vals, a_max)) if a_min <= a_max else a_max
            )

        if isinstance(self.max_vals, np.ndarray):
            max_vals = np.clip(self.max_vals, a_min=a_min, a_max=a_max, *args)
        else:
            max_vals = (
                max(a_min, min(self.max_vals, a_max)) if a_min <= a_max else a_max
            )

        return SingleEntityPhiTensor(
            child=data,
            entity=self.entity,
            min_vals=min_vals,
            max_vals=max_vals,
            scalar_manager=self.scalar_manager,
        )

    def any(
        self,
        axis: Optional[int] = None,
        keepdims: Optional[bool] = False,
        where: Optional[bool] = True,
    ) -> SingleEntityPhiTensor:
        """Test whether any element along a given axis evaluates to True"""

        if is_acceptable_simple_type(self.child):
            if isinstance(self.child, np.ndarray):
                data = self.child.any(
                    axis=axis, out=np.array(True), keepdims=keepdims, where=where
                )
            else:
                # self.child is a singleton
                data = self.child != 0

        elif isinstance(self.child, torch.Tensor):
            data = self.child.numpy().any(
                axis=axis, out=np.array(True), keepdims=keepdims, where=where
            )

        if isinstance(self.min_vals, np.ndarray):
            # test whether any min val evaluates to True
            min_vals = self.min_vals.any(
                axis=axis, out=np.array(True), keepdims=keepdims, where=where
            )
        else:
            min_vals = self.min_vals != 0

        if isinstance(self.max_vals, np.ndarray):
            # test whether any max val evaluates to True
            max_vals = self.max_vals.any(
                axis=axis, out=np.array(True), keepdims=keepdims, where=where
            )
        else:
            max_vals = self.max_vals != 0

        # * 1 just makes sure it returns integers instead of True/False
        return SingleEntityPhiTensor(
            child=data * 1,
            entity=self.entity,
            min_vals=min_vals * 1,
            max_vals=max_vals * 1,
            scalar_manager=self.scalar_manager,
        )

    def all(
        self,
        axis: Optional[int] = None,
        keepdims: Optional[bool] = False,
        where: Optional[bool] = True,
    ) -> SingleEntityPhiTensor:
        """Test whether all elements along a given axis evaluates to True"""

        if is_acceptable_simple_type(self.child):
            if isinstance(self.child, np.ndarray):
                data = self.child.all(
                    axis=axis, out=np.array(True), keepdims=keepdims, where=where
                )
            else:
                # self.child is a singleton
                data = self.child != 0

        elif isinstance(self.child, torch.Tensor):
            data = self.child.numpy().all(
                axis=axis, out=np.array(True), keepdims=keepdims, where=where
            )

        if isinstance(self.min_vals, np.ndarray):
            # test whether all min vals evaluate to True
            min_vals = self.min_vals.all(
                axis=axis, out=np.array(True), keepdims=keepdims, where=where
            )
        else:
            min_vals = self.min_vals != 0

        if isinstance(self.max_vals, np.ndarray):
            # test whether all max vals evaluate to True
            max_vals = self.max_vals.all(
                axis=axis, out=np.array(True), keepdims=keepdims, where=where
            )
        else:
            max_vals = self.max_vals != 0

        # * 1 just makes sure it returns integers instead of True/False
        return SingleEntityPhiTensor(
            child=data * 1,
            entity=self.entity,
            min_vals=min_vals * 1,
            max_vals=max_vals * 1,
            scalar_manager=self.scalar_manager,
        )

    def abs(
        self,
        out: Optional[np.ndarray] = None,
    ) -> SingleEntityPhiTensor:
        """Calculate the absolute value element-wise"""
        if is_acceptable_simple_type(self.child):
            data = self.child.__abs__()

        elif isinstance(self.child, torch.Tensor):
            data = self.child.numpy().__abs__()

        if isinstance(self.min_vals, np.ndarray):
            min_vals = np.abs(self.min_vals, out)
        else:
            min_vals = abs(self.min_vals)

        if isinstance(self.max_vals, np.ndarray):
            max_vals = np.abs(self.max_vals, out)
        else:
            max_vals = abs(self.max_vals)

        return SingleEntityPhiTensor(
            child=data,
            entity=self.entity,
            min_vals=min_vals,
            max_vals=max_vals,
            scalar_manager=self.scalar_manager,
        )

    def pow(self, value: SupportedChainType) -> SingleEntityPhiTensor:
        """Return elements raised to powers from value, element-wise"""

        if isinstance(value, SingleEntityPhiTensor):

            if self.entity != value.entity:
                return NotImplemented

            data = self.child**value.child

            min_min = self.min_vals**value.min_vals
            min_max = self.min_vals**value.max_vals
            max_min = self.max_vals**value.min_vals
            max_max = self.max_vals**value.max_vals

            min_vals = np.min([min_min, min_max, max_min, max_max], axis=0)  # type: ignore
            max_vals = np.max([min_min, min_max, max_min, max_max], axis=0)  # type: ignore
            entity = self.entity

            return SingleEntityPhiTensor(
                child=data,
                entity=entity,
                min_vals=min_vals,
                max_vals=max_vals,
                scalar_manager=self.scalar_manager,
            )

        elif is_acceptable_simple_type(value):

            data = self.child**value

            min_min = self.min_vals**value
            min_max = self.min_vals**value
            max_min = self.max_vals**value
            max_max = self.max_vals**value

            min_vals = np.min([min_min, min_max, max_min, max_max], axis=0)  # type: ignore
            max_vals = np.max([min_min, min_max, max_min, max_max], axis=0)  # type: ignore
            entity = self.entity

            return SingleEntityPhiTensor(
                child=data,
                entity=entity,
                min_vals=min_vals,
                max_vals=max_vals,
                scalar_manager=self.scalar_manager,
            )

        else:
            return NotImplemented

    def copy(
        self, order: Optional[str] = "K", subok: Optional[bool] = True
    ) -> SingleEntityPhiTensor:
        """Return copy of the given object"""
        if is_acceptable_simple_type(self.child):
            data = self.child.copy()

        elif isinstance(self.child, torch.Tensor):
            data = self.child.numpy().copy()

        elif isinstance(self.child, ShareTensor):
            data = self.child.copy()

        if isinstance(self.min_vals, np.ndarray):
            min_vals = np.array(self.min_vals, order=order, subok=subok, copy=True)
        else:
            min_vals = self.min_vals

        if isinstance(self.max_vals, np.ndarray):
            max_vals = np.array(self.max_vals, order=order, subok=subok, copy=True)
        else:
            max_vals = self.max_vals

        return SingleEntityPhiTensor(
            child=data,
            entity=self.entity,
            min_vals=min_vals,
            max_vals=max_vals,
            scalar_manager=self.scalar_manager,
        )

    @staticmethod
    def sept_like(tensor: SingleEntityPhiTensor, ent: Entity) -> SingleEntityPhiTensor:
        """Create a SEPT with identical data, but belonging to a different entity"""
        return SingleEntityPhiTensor(
            child=tensor.child,
            entity=ent,
            min_vals=tensor.min_vals,
            max_vals=tensor.max_vals,
            scalar_manager=tensor.scalar_manager,
        )

    @staticmethod
    def zeros_like(
        tensor: SingleEntityPhiTensor, ent: Optional[Entity] = None
    ) -> SingleEntityPhiTensor:
        """Create a SEPT of the same shape, but with zeros as its data.
        By default, the tensor will have the same entity, but a different entity can be passed instead.
        """
        if ent:
            return SingleEntityPhiTensor(
                child=np.zeros_like(tensor.child),
                entity=ent,
                min_vals=np.zeros_like(tensor.min_vals),
                max_vals=np.ones_like(tensor.max_vals),
                scalar_manager=tensor.scalar_manager,
            )
        else:
            return SingleEntityPhiTensor(
                child=np.zeros_like(tensor.child),
                entity=tensor.entity,
                min_vals=np.zeros_like(tensor.min_vals),
                max_vals=np.ones_like(tensor.max_vals),
                scalar_manager=tensor.scalar_manager,
            )

    @staticmethod
    def ones_like(
        tensor: SingleEntityPhiTensor, ent: Optional[Entity] = None
    ) -> SingleEntityPhiTensor:
        """Create a SEPT of the same shape, but with ones as its data
        By default, the tensor will have the same entity, but a different entity can be passed instead.
        """
        if ent:
            return SingleEntityPhiTensor(
                child=np.ones_like(tensor.child),
                entity=ent,
                min_vals=np.zeros_like(tensor.min_vals),
                max_vals=np.ones_like(tensor.max_vals),
                scalar_manager=tensor.scalar_manager,
            )
        else:
            return SingleEntityPhiTensor(
                child=np.ones_like(tensor.child),
                entity=tensor.entity,
                min_vals=np.zeros_like(tensor.min_vals),
                max_vals=np.ones_like(tensor.max_vals),
                scalar_manager=tensor.scalar_manager,
            )

    def take(
        self,
        indices: np.ArrayLike,
        axis: Optional[int] = None,
        mode: Optional[str] = "raise",
    ) -> SingleEntityPhiTensor:
        """Take elements from an array along an axis"""
        if is_acceptable_simple_type(self.child):
            data = self.child.take(indices=indices, axis=axis, mode=mode)

        elif isinstance(self.child, torch.Tensor):
            data = self.child.numpy().take(indices=indices, axis=axis, mode=mode)

        min_vals = self.min_vals.take(indices=indices, axis=axis, mode=mode)
        max_vals = self.max_vals.take(indices=indices, axis=axis, mode=mode)

        return SingleEntityPhiTensor(
            child=data,
            entity=self.entity,
            min_vals=min_vals,
            max_vals=max_vals,
            scalar_manager=self.scalar_manager,
        )

    def diagonal(
        self,
        offset: Optional[int] = 0,
        axis1: Optional[int] = 0,
        axis2: Optional[int] = 1,
    ) -> SingleEntityPhiTensor:
        """Return specified diagonals"""
        if is_acceptable_simple_type(self.child):
            if (
                isinstance(self.child, int)
                or isinstance(self.child, float)
                or isinstance(self.child, bool)
            ):
                raise Exception(
                    "ValueError: diag requires an array of at least two dimensions"
                )

            elif isinstance(self.child, np.matrix):
                # Make diagonal of matrix 1-D to preserve backward compatibility.
                data = np.asarray(self.child).diagonal(
                    offset=offset, axis1=axis1, axis2=axis2
                )
            else:
                data = np.asanyarray(self.child).diagonal(
                    offset=offset, axis1=axis1, axis2=axis2
                )
        elif isinstance(self.child, torch.Tensor):
            if isinstance(self.child.numpy(), np.matrix):
                # Make diagonal of matrix 1-D to preserve backward compatibility.
                data = np.asarray(self.child.numpy()).diagonal(
                    offset=offset, axis1=axis1, axis2=axis2
                )
            else:
                data = np.asanyarray(self.child.numpy()).diagonal(
                    offset=offset, axis1=axis1, axis2=axis2
                )

        if (
            isinstance(self.min_vals, int)
            or isinstance(self.min_vals, float)
            or isinstance(self.min_vals, bool)
        ):
            raise Exception(
                "ValueError: diag requires an array of at least two dimensions"
            )

        elif isinstance(self.min_vals, np.matrix):
            # Make diagonal of matrix 1-D to preserve backward compatibility.
            min_vals = np.asarray(self.min_vals).diagonal(
                offset=offset, axis1=axis1, axis2=axis2
            )
        else:
            min_vals = np.asanyarray(self.min_vals).diagonal(
                offset=offset, axis1=axis1, axis2=axis2
            )

        if (
            isinstance(self.max_vals, int)
            or isinstance(self.max_vals, float)
            or isinstance(self.max_vals, bool)
        ):
            raise Exception(
                "ValueError: diag requires an array of at least two dimensions"
            )

        elif isinstance(self.max_vals, np.matrix):
            # Make diagonal of matrix 1-D to preserve backward compatibility.
            max_vals = np.asarray(self.max_vals).diagonal(
                offset=offset, axis1=axis1, axis2=axis2
            )
        else:
            max_vals = np.asanyarray(self.max_vals).diagonal(
                offset=offset, axis1=axis1, axis2=axis2
            )

        return SingleEntityPhiTensor(
            child=data,
            entity=self.entity,
            min_vals=min_vals,
            max_vals=max_vals,
            scalar_manager=self.scalar_manager,
        )

    def max(
        self,
        axis: Optional[int] = None,
        out: Optional[np.ndarray] = None,
        keepdims: Optional[bool] = False,
        initial: Optional[int] = None,
        where: bool = True,
    ) -> SingleEntityPhiTensor:
        # Note: Who knew this method had SO MANY ARGUMENTS?!?!?
        if is_acceptable_simple_type(self.child):
            if isinstance(self.child, np.ndarray):
                data = self.child.max(axis, out, keepdims, initial, where)
            else:
                # This implies self.child is a singleton (int, float, bool, etc)
                data = self.child
        elif isinstance(self.child, torch.Tensor):
            data = self.child.numpy().max(axis, out, keepdims, initial, where)
        else:
            raise NotImplementedError

        if isinstance(self.min_vals, np.ndarray):
            min_vals = self.min_vals.max(axis, out, keepdims, initial, where)
        else:
            min_vals = self.min_vals

        if isinstance(self.max_vals, np.ndarray):
            max_vals = self.max_vals.max(axis, out, keepdims, initial, where)
        else:
            max_vals = self.max_vals

        return SingleEntityPhiTensor(
            child=data,
            max_vals=max_vals,
            min_vals=min_vals,
            entity=self.entity,
        )

    def min(
        self,
        axis: Optional[int] = None,
        out: Optional[np.ndarray] = None,
        keepdims: Optional[bool] = False,
        initial: Optional[int] = None,
        where: bool = True,
    ) -> SingleEntityPhiTensor:
        # Note: Who knew this method had SO MANY ARGUMENTS?!?!?
        if is_acceptable_simple_type(self.child):
            if isinstance(self.child, np.ndarray):
                data = self.child.min(axis, out, keepdims, initial, where)
            else:
                # This implies self.child is a singleton (int, float, bool, etc)
                data = self.child
        elif isinstance(self.child, torch.Tensor):
            data = self.child.numpy().min(axis, out, keepdims, initial, where)
        else:
            raise NotImplementedError

        if isinstance(self.min_vals, np.ndarray):
            min_vals = self.min_vals.min(axis, out, keepdims, initial, where)
        else:
            min_vals = self.min_vals

        if isinstance(self.max_vals, np.ndarray):
            max_vals = self.max_vals.min(axis, out, keepdims, initial, where)
        else:
            max_vals = self.max_vals

        return SingleEntityPhiTensor(
            child=data,
            max_vals=max_vals,
            min_vals=min_vals,
            entity=self.entity,
        )

    # TODO: Figure out how to do type annotation for dtype
    def trace(
        self,
        offset: int = 0,
        axis1: Optional[int] = 0,
        axis2: Optional[int] = 1,
        dtype: Optional[Any] = None,
        out: np.ndarray = None,
    ) -> SingleEntityPhiTensor:
        if isinstance(self.child, np.ndarray):
            data = self.child.trace(offset, axis1, axis2, dtype, out)
        elif isinstance(self.child, torch.Tensor):
            data = self.child.numpy().trace(offset, axis1, axis2, dtype, out)
        else:
            data = self.child * len(self.child)

        if isinstance(self.min_vals, np.ndarray):
            mins = self.min_vals.trace(offset, axis1, axis2, dtype, out)
        else:
            mins = self.min_vals * len(self.child)

        if isinstance(self.max_vals, np.ndarray):
            maxes = self.max_vals.trace(offset, axis1, axis2, dtype, out)
        else:
            maxes = self.max_vals * len(self.child)

        return SingleEntityPhiTensor(
            child=data,
            min_vals=mins,
            max_vals=maxes,
            entity=self.entity,
        )

    def prod(
        self,
        axis: Optional[int] = None,
        dtype: Optional[Any] = None,
        out: Optional[np.ndarray] = None,
        keepdims: Optional[bool] = False,
        initial: int = 1,
        where: Optional[bool] = True,
    ) -> SingleEntityPhiTensor:
        return SingleEntityPhiTensor(
            child=self.child.prod(axis, dtype, out, keepdims, initial, where),
            min_vals=self.min_vals.prod(axis, dtype, out, keepdims, initial, where),
            max_vals=self.max_vals.prod(axis, dtype, out, keepdims, initial, where),
            entity=self.entity,
        )

    def round(self, decimals: int = 0) -> SingleEntityPhiTensor:
        if decimals != 0:
            raise Exception(
                "We currently only support np.int32. Sorry about the inconvenience-"
                "we plan to support more types soon!"
            )
        return SingleEntityPhiTensor(
            child=self.child.astype(dtype=np.int32),
            min_vals=self.min_vals.astype(dtype=np.int32),
            max_vals=self.max_vals.astype(dtype=np.int32),
            scalar_manager=self.scalar_manager,
            entity=self.entity,
        )

    def __floordiv__(
        self, other: Union[AcceptableSimpleType, SingleEntityPhiTensor]
    ) -> Union[SingleEntityPhiTensor, IntermediateGammaTensor]:
        if is_acceptable_simple_type(other):
            if isinstance(other, np.ndarray) and not is_broadcastable(
                self.shape, other.shape
            ):
                raise Exception(
                    f"Shapes not broadcastable: {self.shape} and {other.shape}"
                )
            else:
                data = self.child // other
                mins = self.min_vals // other
                maxes = self.max_vals // other
        elif isinstance(other, SingleEntityPhiTensor):
            if is_broadcastable(self.shape, other.shape):
                if self.entity == other.entity:
                    data = self.child // other.child
                    mins = self.min_vals // other.min_vals
                    maxes = self.max_vals // other.max_vals
                else:
                    # return convert_to_gamma_tensor(self) // convert_to_gamma_tensor(other)
                    raise NotImplementedError
            else:
                raise Exception(
                    f"Shapes not broadcastable: {self.shape} and {other.shape}"
                )
        else:
            raise NotImplementedError
        return SingleEntityPhiTensor(
            child=data,
            max_vals=maxes,
            min_vals=mins,
            entity=self.entity,
            scalar_manager=self.scalar_manager,
        )

    #
    # # TODO: Check to see if non-integers are ever introduced
    # def __mod__(
    #     self, other: Union[AcceptableSimpleType, SingleEntityPhiTensor]
    # ) -> Union[SingleEntityPhiTensor, IntermediateGammaTensor]:
    #     if is_acceptable_simple_type(other):
    #         if isinstance(other, np.ndarray) and not is_broadcastable(
    #             self.shape, other.shape
    #         ):
    #             raise Exception(
    #                 f"Shapes not broadcastable: {self.shape} and {other.shape}"
    #             )
    #         else:
    #             data = self.child % other
    #             mins = self.min_vals % other
    #             maxes = self.max_vals % other
    #     elif isinstance(other, SingleEntityPhiTensor):
    #         if is_broadcastable(self.shape, other.shape):
    #             if self.entity == other.entity:
    #                 data = self.child % other.child
    #                 mins = self.min_vals % other.min_vals
    #                 maxes = self.max_vals % other.max_vals
    #             else:
    #                 # return convert_to_gamma_tensor(self) % convert_to_gamma_tensor(other)
    #                 raise NotImplementedError
    #         else:
    #             raise Exception(
    #                 f"Shapes not broadcastable: {self.shape} and {other.shape}"
    #             )
    #     else:
    #         raise NotImplementedError
    #     return SingleEntityPhiTensor(
    #         child=data,
    #         max_vals=maxes,
    #         min_vals=mins,
    #         entity=self.entity,
    #         scalar_manager=self.scalar_manager,
    #     )
    # #
    # # def __divmod__(
    # #     self, other: Union[AcceptableSimpleType, SingleEntityPhiTensor]
    # # ) -> TypeTuple:
    # #     return self // other, self % other

    def __matmul__(
        self, other: Union[np.ndarray, SingleEntityPhiTensor]
    ) -> Union[SingleEntityPhiTensor, IntermediateGammaTensor]:
        if not isinstance(other, (np.ndarray, SingleEntityPhiTensor)):
            raise Exception(
                f"Matrix multiplication not yet implemented for type {type(other)}"
            )
        else:
            if not is_broadcastable(self.shape, other.shape):
                raise Exception(
                    f"Shapes not broadcastable: {self.shape} and {other.shape}"
                )
            else:
                if isinstance(other, np.ndarray):
                    data = self.child.__matmul__(other)
                    mins = self.min_vals.__matmul__(other)
                    maxes = self.max_vals.__matmul__(other)
                elif isinstance(other, SingleEntityPhiTensor):
                    if self.entity != other.entity:
                        # return convert_to_gamma_tensor(self).__matmul__(convert_to_gamma_tensor(other))
                        raise NotImplementedError
                    else:
                        data = self.child.__matmul__(other.child)
                        mins = self.min_vals.__matmul__(other.min_vals)
                        maxes = self.max_vals.__matmul__(other.max_vals)
                else:
                    raise NotImplementedError
                return SingleEntityPhiTensor(
                    child=data,
                    max_vals=maxes,
                    min_vals=mins,
                    entity=self.entity,
                    scalar_manager=self.scalar_manager,
                )

    def cumsum(
        self,
        axis: Optional[int] = None,
        dtype: Optional[Any] = None,
        out: np.ndarray = None,
    ) -> SingleEntityPhiTensor:
        if dtype and dtype != np.int32:
            raise Exception(
                "We currently only support np.int32 dtypes. "
                "We have plans to support more in the future though!"
            )
        if isinstance(self.child, np.ndarray):
            data = self.child.cumsum(axis, dtype, out)
        elif isinstance(self.child, torch.Tensor):
            data = self.child.numpy().cumsum(axis, dtype, out)
        else:
            data = self.child * len(self.child)

        if isinstance(self.min_vals, np.ndarray):
            mins = self.min_vals.cumsum(axis, dtype, out)
        else:
            mins = self.min_vals * len(self.child)

        if isinstance(self.max_vals, np.ndarray):
            maxes = self.max_vals.cumsum(axis, dtype, out)
        else:
            maxes = self.max_vals * len(self.child)

        return SingleEntityPhiTensor(
            child=data, min_vals=mins, max_vals=maxes, entity=self.entity
        )

    def cumprod(
        self,
        axis: Optional[int] = None,
        dtype: Optional[Any] = None,
        out: np.ndarray = None,
    ) -> SingleEntityPhiTensor:
        if dtype and dtype != np.int32:
            raise Exception(
                "We currently only support np.int32 dtypes. "
                "We have plans to support more in the future though!"
            )
        if isinstance(self.child, np.ndarray):
            data = self.child.cumprod(axis, dtype, out)
        elif isinstance(self.child, torch.Tensor):
            data = self.child.numpy().cumprod(axis, dtype, out)
        else:
            data = self.child * len(self.child)

        if isinstance(self.min_vals, np.ndarray):
            mins = self.min_vals.cumprod(axis, dtype, out)
        else:
            mins = self.min_vals * len(self.child)

        if isinstance(self.max_vals, np.ndarray):
            maxes = self.max_vals.cumprod(axis, dtype, out)
        else:
            maxes = self.max_vals * len(self.child)

        return SingleEntityPhiTensor(
            child=data, min_vals=mins, max_vals=maxes, entity=self.entity
        )

    def _object2proto(self) -> SingleEntityPhiTensor_PB:
        proto_init_kwargs = {
            "min_vals": serialize(self._min_vals),
            "max_vals": serialize(self._max_vals),
        }

        # to de-duplicate entity and scalar manager at the RowEntityPhiTensor level
        # we need to allow this to sometimes be None and not store it
        if not getattr(self, "_remove_entity_scalar_manager", False):
            proto_init_kwargs["entity"] = serialize(self.entity)
            proto_init_kwargs["scalar_manager"] = serialize(self.scalar_manager)

        # either numpy array or ShareTensor
        if self.child is None:
            proto_init_kwargs["none"] = serialize(self.child)
        elif isinstance(self.child, np.ndarray):
            proto_init_kwargs["array"] = serialize(self.child)
        elif isinstance(self.child, torch.Tensor):
            proto_init_kwargs["array"] = serialize(np.array(self.child))
        else:
            proto_init_kwargs["tensor"] = serialize(self.child)

        return SingleEntityPhiTensor_PB(**proto_init_kwargs)

    @staticmethod
    def _proto2object(proto: SingleEntityPhiTensor_PB) -> SingleEntityPhiTensor:
        # either numpy array or ShareTensor
        if proto.HasField("tensor"):
            child = deserialize(proto.tensor)
        elif proto.HasField("array"):
            child = deserialize(proto.array)
        else:
            child = deserialize(proto.none)

        return SingleEntityPhiTensor(
            child=child,
            entity=deserialize(proto.entity) if proto.HasField("entity") else None,
            min_vals=deserialize(proto.min_vals),
            max_vals=deserialize(proto.max_vals),
            scalar_manager=deserialize(proto.scalar_manager)
            if proto.HasField("scalar_manager")
            else None,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return SingleEntityPhiTensor_PB


@implements(SingleEntityPhiTensor, np.expand_dims)
def expand_dims(a: npt.ArrayLike, axis: Optional[int] = None) -> SingleEntityPhiTensor:

    entity = a.entity  # type: ignore

    min_vals = np.expand_dims(a=a.min_vals, axis=axis)  # type: ignore
    max_vals = np.expand_dims(a=a.max_vals, axis=axis)  # type: ignore

    data = np.expand_dims(a.child, axis=axis)  # type: ignore

    return SingleEntityPhiTensor(
        child=data,
        entity=entity,
        min_vals=min_vals,
        max_vals=max_vals,
        scalar_manager=a.scalar_manager,  # type: ignore
    )


@implements(SingleEntityPhiTensor, np.mean)
def mean(*args: Any, **kwargs: Any) -> SingleEntityPhiTensor:
    entity = args[0].entity
    scalar_manager = args[0].scalar_manager

    for arg in args[1:]:
        if not isinstance(arg, SingleEntityPhiTensor):
            raise Exception("Can only call np.mean on objects of the same type.")

        if arg.entity != entity:
            return NotImplemented

    min_vals = np.mean([x.min_vals for x in args], **kwargs)
    max_vals = np.mean([x.max_vals for x in args], **kwargs)

    args, kwargs = inputs2child(*args, **kwargs)  # type: ignore

    data = np.mean(args, **kwargs)

    return SingleEntityPhiTensor(
        child=data,
        entity=entity,
        min_vals=min_vals,
        max_vals=max_vals,
        scalar_manager=scalar_manager,
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/adp/scalar.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.adp import entity_pb2 as proto_dot_core_dot_adp_dot_entity__pb2
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1bproto/core/adp/scalar.proto\x12\rsyft.core.adp\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/adp/entity.proto"7\n\x12IntermediateScalar\x12!\n\x02id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID"\xb8\x01\n\nBaseScalar\x12!\n\x02id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x14\n\x07min_val\x18\x02 \x01(\x01H\x00\x88\x01\x01\x12\x12\n\x05value\x18\x03 \x01(\x01H\x01\x88\x01\x01\x12\x14\n\x07max_val\x18\x04 \x01(\x01H\x02\x88\x01\x01\x12%\n\x06\x65ntity\x18\x05 \x01(\x0b\x32\x15.syft.core.adp.EntityB\n\n\x08_min_valB\x08\n\x06_valueB\n\n\x08_max_val"<\n\x17IntermediateGammaScalar\x12!\n\x02id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID"\x89\x02\n\x0bGammaScalar\x12!\n\x02id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x14\n\x07min_val\x18\x02 \x01(\x01H\x01\x88\x01\x01\x12\x12\n\x05value\x18\x03 \x01(\x01H\x02\x88\x01\x01\x12\x14\n\x07max_val\x18\x04 \x01(\x01H\x03\x88\x01\x01\x12\'\n\x06\x65ntity\x18\x05 \x01(\x0b\x32\x15.syft.core.adp.EntityH\x00\x12.\n\x03\x64sg\x18\x06 \x01(\x0b\x32\x1f.syft.core.adp.DataSubjectGroupH\x00\x12\r\n\x05prime\x18\x07 \x01(\x03\x42\r\n\x0b\x65ntity_dataB\n\n\x08_min_valB\x08\n\x06_valueB\n\n\x08_max_val"\x9b\x01\n\x15IntermediatePhiScalar\x12!\n\x02id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12%\n\x06\x65ntity\x18\x02 \x01(\x0b\x32\x15.syft.core.adp.Entity\x12.\n\x05gamma\x18\x03 \x01(\x0b\x32\x1a.syft.core.adp.GammaScalarH\x00\x88\x01\x01\x42\x08\n\x06_gamma"\xf1\x01\n\tPhiScalar\x12!\n\x02id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x14\n\x07min_val\x18\x02 \x01(\x01H\x00\x88\x01\x01\x12\x12\n\x05value\x18\x03 \x01(\x01H\x01\x88\x01\x01\x12\x14\n\x07max_val\x18\x04 \x01(\x01H\x02\x88\x01\x01\x12%\n\x06\x65ntity\x18\x05 \x01(\x0b\x32\x15.syft.core.adp.Entity\x12.\n\x05gamma\x18\x06 \x01(\x0b\x32\x1a.syft.core.adp.GammaScalarH\x03\x88\x01\x01\x42\n\n\x08_min_valB\x08\n\x06_valueB\n\n\x08_max_valB\x08\n\x06_gammab\x06proto3'
)


_INTERMEDIATESCALAR = DESCRIPTOR.message_types_by_name["IntermediateScalar"]
_BASESCALAR = DESCRIPTOR.message_types_by_name["BaseScalar"]
_INTERMEDIATEGAMMASCALAR = DESCRIPTOR.message_types_by_name["IntermediateGammaScalar"]
_GAMMASCALAR = DESCRIPTOR.message_types_by_name["GammaScalar"]
_INTERMEDIATEPHISCALAR = DESCRIPTOR.message_types_by_name["IntermediatePhiScalar"]
_PHISCALAR = DESCRIPTOR.message_types_by_name["PhiScalar"]
IntermediateScalar = _reflection.GeneratedProtocolMessageType(
    "IntermediateScalar",
    (_message.Message,),
    {
        "DESCRIPTOR": _INTERMEDIATESCALAR,
        "__module__": "proto.core.adp.scalar_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.adp.IntermediateScalar)
    },
)
_sym_db.RegisterMessage(IntermediateScalar)

BaseScalar = _reflection.GeneratedProtocolMessageType(
    "BaseScalar",
    (_message.Message,),
    {
        "DESCRIPTOR": _BASESCALAR,
        "__module__": "proto.core.adp.scalar_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.adp.BaseScalar)
    },
)
_sym_db.RegisterMessage(BaseScalar)

IntermediateGammaScalar = _reflection.GeneratedProtocolMessageType(
    "IntermediateGammaScalar",
    (_message.Message,),
    {
        "DESCRIPTOR": _INTERMEDIATEGAMMASCALAR,
        "__module__": "proto.core.adp.scalar_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.adp.IntermediateGammaScalar)
    },
)
_sym_db.RegisterMessage(IntermediateGammaScalar)

GammaScalar = _reflection.GeneratedProtocolMessageType(
    "GammaScalar",
    (_message.Message,),
    {
        "DESCRIPTOR": _GAMMASCALAR,
        "__module__": "proto.core.adp.scalar_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.adp.GammaScalar)
    },
)
_sym_db.RegisterMessage(GammaScalar)

IntermediatePhiScalar = _reflection.GeneratedProtocolMessageType(
    "IntermediatePhiScalar",
    (_message.Message,),
    {
        "DESCRIPTOR": _INTERMEDIATEPHISCALAR,
        "__module__": "proto.core.adp.scalar_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.adp.IntermediatePhiScalar)
    },
)
_sym_db.RegisterMessage(IntermediatePhiScalar)

PhiScalar = _reflection.GeneratedProtocolMessageType(
    "PhiScalar",
    (_message.Message,),
    {
        "DESCRIPTOR": _PHISCALAR,
        "__module__": "proto.core.adp.scalar_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.adp.PhiScalar)
    },
)
_sym_db.RegisterMessage(PhiScalar)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _INTERMEDIATESCALAR._serialized_start = 114
    _INTERMEDIATESCALAR._serialized_end = 169
    _BASESCALAR._serialized_start = 172
    _BASESCALAR._serialized_end = 356
    _INTERMEDIATEGAMMASCALAR._serialized_start = 358
    _INTERMEDIATEGAMMASCALAR._serialized_end = 418
    _GAMMASCALAR._serialized_start = 421
    _GAMMASCALAR._serialized_end = 686
    _INTERMEDIATEPHISCALAR._serialized_start = 689
    _INTERMEDIATEPHISCALAR._serialized_end = 844
    _PHISCALAR._serialized_start = 847
    _PHISCALAR._serialized_end = 1088
# @@protoc_insertion_point(module_scope)



# third party
import numpy as np

# relative
# from .single_entity_phi import SingleEntityPhiTensor as SEPT
from .initial_gamma import InitialGammaTensor


# x is a SingleEntityPhiTensor, but we can't import it here since it creates a circular import error.
def convert_to_gamma_tensor(x) -> InitialGammaTensor:  # type: ignore
    """Helper function to convert (x) SEPT to InitialGammaTensor, for private-private operations"""
    return InitialGammaTensor(
        values=x.child,
        min_vals=x.min_vals,
        max_vals=x.max_vals,
        entities=np.array(x.entity, dtype=object).repeat(len(x.child.flatten())),
        scalar_manager=x.scalar_manager,
    )



# third party
import numpy as np

# relative
from ..util import implements
from ..util import inputs2child
from .tensor import AutogradTensor


@implements(AutogradTensor, np.max)
def npmax(*args: AutogradTensor, **kwargs: AutogradTensor) -> AutogradTensor:
    args, kwargs = inputs2child(*args, **kwargs)
    return np.max(*args, **kwargs)


@implements(AutogradTensor, np.min)
def npmin(*args: AutogradTensor, **kwargs: AutogradTensor) -> AutogradTensor:
    args, kwargs = inputs2child(*args, **kwargs)
    return np.min(*args, **kwargs)


@implements(AutogradTensor, np.expand_dims)
def expand_dims(*args: AutogradTensor, **kwargs: AutogradTensor) -> AutogradTensor:
    requires_grad = args[0].requires_grad
    output_type = args[0].__class__
    args, kwargs = inputs2child(*args, **kwargs)
    return output_type(np.expand_dims(*args, **kwargs), requires_grad=requires_grad)


@implements(AutogradTensor, np.multiply)
def multiply(a: AutogradTensor, b: AutogradTensor) -> AutogradTensor:
    return AutogradTensor(a * b)



# future
from __future__ import annotations

# stdlib
import textwrap
from typing import Any
from typing import List
from typing import Optional
from typing import Type
import uuid

# third party
from nacl.signing import VerifyKey
import numpy as np
from numpy.typing import ArrayLike

# relative
from ..adp.entity import Entity
from ..adp.entity_list import EntityList
from ..adp.vm_private_scalar_manager import VirtualMachinePrivateScalarManager
from .lazy_repeat_array import lazyrepeatarray
from .manager import TensorChainManager
from .passthrough import PassthroughTensor  # type: ignore
from .passthrough import is_acceptable_simple_type  # type: ignore

_SingleEntityPhiTensorRef = None


def _SingleEntityPhiTensor() -> Type[PassthroughTensor]:
    global _SingleEntityPhiTensorRef
    if _SingleEntityPhiTensorRef is None:
        # relative
        from .autodp.single_entity_phi import SingleEntityPhiTensor

        _SingleEntityPhiTensorRef = SingleEntityPhiTensor
    return _SingleEntityPhiTensorRef


_RowEntityPhiTensorRef = None


def _RowEntityPhiTensor() -> Type[PassthroughTensor]:
    global _RowEntityPhiTensorRef
    if _RowEntityPhiTensorRef is None:
        # relative
        from .autodp.row_entity_phi import RowEntityPhiTensor

        _RowEntityPhiTensorRef = RowEntityPhiTensor
    return _RowEntityPhiTensorRef


_NDimEntityPhiTensorRef = None


def _NDimEntityPhiTensor() -> Type[PassthroughTensor]:
    global _NDimEntityPhiTensorRef
    if _NDimEntityPhiTensorRef is None:
        # relative
        from .autodp.ndim_entity_phi import NDimEntityPhiTensor

        _NDimEntityPhiTensorRef = NDimEntityPhiTensor
    return _NDimEntityPhiTensorRef


_AutogradTensorRef = None


def _AutogradTensor() -> Type[PassthroughTensor]:
    global _AutogradTensorRef
    if _AutogradTensorRef is None:
        # relative
        from .autograd.tensor import AutogradTensor

        _AutogradTensorRef = AutogradTensor
    return _AutogradTensorRef


class AutogradTensorAncestor(TensorChainManager):
    """Inherited by any class which might have or like to have AutogradTensor in its chain
    of .child objects"""

    @property
    def grad(self):  # type: ignore
        child_gradient = self.child.grad
        if child_gradient is None:
            return None
        return self.__class__(child_gradient)

    @property
    def requires_grad(self) -> bool:
        return self.child.requires_grad

    def backward(self, grad=None):  # type: ignore

        AutogradTensor = _AutogradTensor()

        # TODO: @Madhava question, if autograd(requires_grad=True) is not set
        # we still end up in here from AutogradTensorAncestor but child.backward
        # has no backprop_id
        if isinstance(self.child, AutogradTensorAncestor) or isinstance(
            self.child, AutogradTensor
        ):

            if grad is not None and not is_acceptable_simple_type(grad):
                grad = grad.child

            return self.child.backward(grad, backprop_id=uuid.uuid4())  # type: ignore
        else:
            raise Exception(
                "No AutogradTensor found in chain, but backward() method called."
            )

    def autograd(self, requires_grad: bool = True) -> AutogradTensorAncestor:
        AutogradTensor = _AutogradTensor()

        self.push_abstraction_top(AutogradTensor, requires_grad=requires_grad)  # type: ignore

        return self


def entity_creation_wizard(data: Any) -> List[Any]:

    w = textwrap.TextWrapper(initial_indent="\t", subsequent_indent="\t")

    welcome_msg = "Welcome to the Data Subject Annotation Wizard!!!"

    description1 = """You've arrived here because you called Tensor.private() without passing in any entities!
Since the purpose of .private() is to add metadata for the support of automatic differential
privacy budgeting, you need to describe which parts of your Tensor correspond to which
real-world data subjects (entities) whose privacy you want to protect. This is the only
way the system knows, for example, that it costs twice as much privacy budget when twice
as much of your data (say, 2 rows instead of 1 row) refer to the same entity."""

    description2 = """Entities can be people (such as a medical patient), places (such as a family's address), or
even organizations (such as a business, state, or country). If you're not sure what kind of entity
to include, just ask yourself the question, "who am I trying to protect the privacy of?". If it's
an organization, make one entity per organization. If it's people, make one entity per person.
If it's a group of people who are somehow similar/linked to each other (such as a family),
make each entity a different group. For more information on differential privacy, see OpenMined's
course on the subject: https://courses.openmined.org/"""

    description3 = """Since you didn't pass in entities into .private() (or you did so incorrectly), this wizard is
going to guide you through the process of annotating your data with entities."""

    description4 = """In this wizard, we're going to ask you for *unique identifiers* which refer to the entities
in your data. While the unique identifiers need not be personal data (they can be random strings of letters and numbers
if you like). It is ESSENTIAL that you use the same identifier when referring to the same entity in the
data that you never accidentally refer to two entities by the same identifier. Additionally, if you plan
to do any kind of data JOIN with another dataset, it is ESSENTIAL that you are using the same unique
identifiers for entities as the data you're joining with. Since these unique identifiers may be personal
information, PySyft might not be able to detect if two tensors are using different identifiers for the
same person."""

    description5 = """So, in this tutorial we're going to be asking you to specify Unique Identifiers (UIDs) for each entity
in your data. This could be an email, street address, or any other string that identifies someone
uniquely in your data and in the data you intend to use with your data (if any)."""

    print("\t" + "=" * 69)
    print(w.fill(welcome_msg))
    print("\t" + "=" * 69)
    print()
    print(w.fill(description1))
    print()
    print(w.fill(description2))
    print()
    print(w.fill(description3))
    print()
    print(w.fill(description4))
    print()
    print(w.fill(description5))
    print()

    print("\tDo you understand, and are you ready to proceed? (yes/no)")
    print()
    consent = str(input("\t"))
    print()

    if consent == "no":
        raise Exception("User cancelled entity creation wizard!")

    print("\tExcellent! Let's begin!")
    # print("\tYou passed in a tensor with the shape:" + str(data.shape))
    print()

    print("\t" + "-" * 69)
    print()

    print(w.fill("Question 1: Is this entire tensor referring to the same entity?"))
    print()
    print(w.fill("Examples:"))
    print("\t - a single medical scan of one patient")
    print("\t - a single spreadsheet of proprietary statistics about a business")
    print("\t - a tensor of facts about a country")
    print()
    print(
        w.fill(
            """(if the tensor is about one entity, but it also contains multiple other entities within,
such as a tensor about all the customers of one business, ask yourself, are you trying to
protect the people or the business)"""
        )
    )
    print()
    print(
        w.fill(
            "If yes, write the UID of the entity this data is about, otherwise write 'no' "
            " because this data is about more than one entity."
        )
    )
    print()
    single_uid = input("\t")
    print()
    if single_uid != "no":
        print("\t" + "-" * 69)
        print()
        print(
            w.fill(
                "Excellent! Your data will be annotated as referring to:"
                + str(single_uid)
            )
        )
        print()
        print(
            w.fill(
                "Congratulations! You're all done with the Data Subject Annotation Wizard!!!"
                "In the future, you can accomplish this without the wizard by running:"
            )
        )
        print()
        print(w.fill("\t.private(entities='" + str(single_uid) + "')"))
        print()
        print("\t" + "=" * 69)
        return [single_uid]

    print("\t" + "-" * 69)
    print()
    print(
        w.fill(
            "Question 2: Does each row correspond to an entity, perhaps with occasional repeats (yes/no)?"
        )
    )
    print()
    answer = str(input("\t"))
    print()
    print("\t" + "-" * 69)
    print()
    if answer == "yes":
        print(
            w.fill(
                "Question 3: Excellent! Well, since your dataset has "
                + str(data.shape[0])
                + " rows, "
                + "would you like to hand enter an entity for each one (yes) or if there are too "
                + "many for you to hand-enter, we'll print some example code for you to run (no)."
            )
        )

        print()

        answer = str(input("\t"))

        if answer == "yes":

            print()

            entities = list()
            for i in range(len(data)):
                print("\t\t" + "-" * 61)
                print()
                print(w.fill("\tData Row " + str(i) + ":" + str(data[i])))
                ent = input("\t\t What entity is this row about:")
                entities.append(ent)
                print()
            print("\t\t" + "-" * 61)
            print()
            print(
                w.fill(
                    "All done! Next time if you want to skip the wizard, call .private() like this:"
                )
            )
            print()
            print(
                w.fill(
                    ".private(entities=['"
                    + entities[0]
                    + "', '"
                    + entities[1]
                    + "', '"
                    + entities[-1]
                    + "'])"
                )
            )
            print()
            print(
                w.fill(
                    " where you pass in entities as a list of strings, one per row. As long as you"
                    " pass in the same number of entities as there are rows in your tensor, it will"
                    " automatically detect you have and assume you mean one entity per row."
                )
            )
            return entities

        elif answer == "no":

            print()

            print(
                w.fill(
                    "Excellent. Well, in that case you'll need to re-run .private() but pass in"
                    " a list of strings where each string is a unique identifier for an entity, and where"
                    " the length of the list is equal to the number of rows in your tensor. Like so:"
                )
            )

            print()
            print(w.fill(".private(entities=['bob', 'alice', 'john'])"))
            print()
            print(
                " Now just to make sure I don't corrupt your tensor - I'm going to throw an exception."
            )
            print()
            raise Exception(
                "Wizard aborted. Please run .private(entities=<your entities>)"
                " again with your list of entity unique identifiers (strings),"
                "one per row of your tensor."
            )
    elif answer == "no":

        print(w.fill("Question 3: Is your data one entity for every column (yes/no)?"))

        print()

        answer = str(input("\t"))

        print()

        if answer == "yes":
            print(
                w.fill(
                    "We don't yet support this form of injestion. Please transpose your data"
                    " into one entity per row and re-run the wizard. Aborting:)"
                )
            )

            raise Exception("Wizard aborted.")

        elif answer == "no":

            print(
                w.fill(
                    "It sounds like your tensor is a random assortment of entities (and perhaps empty/non-entities). "
                    "If you have empty values, just create random entities for them for now. If you have various "
                    "entities scattered throughout your tensor (not organized by row), then you'll need to pass "
                    "in a np.ndarray of strings which is identically shaped to your data in entities like so:"
                )
            )

            print()
            print("\t\ttensor = sy.Tensor(np.ones((2,2)).astype(np.int32))")
            print()
            print("\t\tentities = np.array([['bob', 'alice'],['charlie', 'danielle']])")
            print()
            print("\t\ttensor.private(min_val=0, max_val=1, entities=entities))")
            print()
            print(
                "Aborting wizard now so that you rcan re-run .private with the right parameters."
            )
            print()
            raise Exception(
                "Wizard aborted. Please run .private(entities=<your entities>)"
                " again with your np.ndarray of entity unique identifiers (strings),"
                " one per value of your tensor and where your np.ndarray of entities is"
                " the same shape as your data."
            )
    print()

    print("\t" + "_" * 69)
    raise Exception(
        "Not sure what happened... this code shouldn't have been reached. Try answering questions with "
        "options given by the prompts (such as yes/no)."
    )


class PhiTensorAncestor(TensorChainManager):
    """Inherited by any class which might have or like to have SingleEntityPhiTensor in its chain
    of .child objects"""

    def __init__(self, child: Any) -> None:
        self.child = child

    @property
    def shape(self) -> List[int]:
        return self.child.shape

    @property
    def min_vals(self):  # type: ignore
        return self.__class__(self.child.min_vals)

    @property
    def max_vals(self):  # type: ignore
        return self.__class__(self.child.max_vals)

    @property
    def gamma(self):  # type: ignore
        return self.__class__(self.child.gamma)

    def publish(self, acc: Any, sigma: float, user_key: VerifyKey) -> PhiTensorAncestor:
        return self.child.publish(acc=acc, sigma=sigma, user_key=user_key)

    def copy(self) -> PhiTensorAncestor:
        """This should certainly be implemented by the subclass but adding this here to satisfy mypy."""

        return NotImplemented

    def private(
        self,
        min_val: ArrayLike,
        max_val: ArrayLike,
        scalar_manager: VirtualMachinePrivateScalarManager = VirtualMachinePrivateScalarManager(),
        entities: Optional[Any] = None,
        skip_blocking_checks: bool = False,
        ndept: bool = False,
    ) -> PhiTensorAncestor:
        return self.copy()._private(
            min_val=min_val,
            max_val=max_val,
            scalar_manager=scalar_manager,
            entities=entities,
            skip_blocking_checks=skip_blocking_checks,
            ndept=ndept,
        )

    def _private(
        self,
        min_val: ArrayLike,
        max_val: ArrayLike,
        scalar_manager: VirtualMachinePrivateScalarManager = VirtualMachinePrivateScalarManager(),
        entities: Optional[Any] = None,
        skip_blocking_checks: bool = False,
        ndept: bool = False,
    ) -> PhiTensorAncestor:
        # PHASE 1: RUN CHECKS

        # Check 1: Is self.child a compatible type? We only support DP and SMPC for a few types.
        if (
            not isinstance(self.child, np.ndarray)
            or getattr(self.child, "dtype", None) != np.int32
        ):

            msg = (
                "At present, you can only call .private() "
                + "on syft.Tensor objects wrapping np.int32 arrays. You called it on a "
                + "syft.Tensor wrapping a "
                + str(type(self.child))
            )

            if isinstance(self.child, np.ndarray):
                msg += " with dtype:" + str(getattr(self.child, "dtype", None))

            raise TypeError(msg)

        # Check 2: If entities == None, then run the entity creation tutorial
        if entities is None:
            if skip_blocking_checks:
                raise Exception(
                    "Error: 'entities' argument to .private() must not be None!"
                )
            print(
                "ALERT: You didn't pass in any entities. Launching entity wizard...\n"
            )
            entities = entity_creation_wizard(self.child)

        # Check 3: If entities is a string, make it a list with one entity in it
        if isinstance(entities, str):
            entities = [Entity(entities)]
        elif isinstance(entities, Entity):
            entities = [entities]
        # Check 4: If entities are a list, are the items strings or Entity objects.
        # If they're strings lets create Entity objects.

        if isinstance(entities, (list, tuple)):
            entities = np.array(entities)

        # if len(entities) != 1 and entities.shape != self.shape:
        #     raise Exception(
        #         "Entities shape doesn't match data shape. If you're"
        #         " going to pass in something other than 1 entity for the"
        #         " entire tensor or one entity per row, you're going to need"
        #         " to make the np.ndarray of entities have the same shape as"
        #         " the tensor you're calling .private() on. Try again."
        #     )

        if not isinstance(entities, EntityList):
            one_hot_lookup, entities_indexed = np.unique(entities, return_inverse=True)
        else:
            one_hot_lookup, entities_indexed = (
                entities.one_hot_lookup,
                entities.entities_indexed,
            )

        for entity in one_hot_lookup:
            if not isinstance(entity, (str, Entity)):
                raise ValueError(
                    f"Expected Entity to be either string or Entity object, but type is {type(entity)}"
                )

        # PHASE 2: CREATE CHILD
        if len(entities) == 1:
            # if there's only one entity - push a SingleEntityPhiTensor
            if isinstance(min_val, (float, int)):
                min_vals = (self.child * 0) + min_val
            else:
                raise Exception(
                    "min_val should be a float, got " + str(type(min_val)) + " instead."
                )

            if isinstance(max_val, (float, int)):
                max_vals = (self.child * 0) + max_val
            else:
                raise Exception(
                    "max_val should be a float, got " + str(type(min_val)) + " instead."
                )

            self.push_abstraction_top(
                _SingleEntityPhiTensor(),
                entity=entities[0],  # type: ignore
                min_vals=min_vals,
                max_vals=max_vals,
                scalar_manager=scalar_manager,  # type: ignore
            )

        # if there's row-level entities - push a RowEntityPhiTensor
        elif not ndept and entities is not None and len(entities) == self.shape[0]:
            class_type = _SingleEntityPhiTensor()

            new_list = list()
            for i, entity in enumerate(entities):

                if isinstance(min_val, (float, int)):
                    min_vals = (self.child[i : i + 1] * 0) + min_val  # noqa: E203
                else:
                    raise Exception(
                        "min_val should be a float, got "
                        + str(type(min_val))
                        + " instead."
                    )

                if isinstance(max_val, (float, int)):
                    max_vals = (self.child[i : i + 1] * 0) + max_val  # noqa: E203
                else:
                    raise Exception(
                        "max_val should be a float, got "
                        + str(type(min_val))
                        + " instead."
                    )

                value = self.child[i : i + 1]  # noqa: E203

                new_list.append(
                    class_type(
                        child=value,
                        entity=entity,
                        min_vals=min_vals,
                        max_vals=max_vals,
                        scalar_manager=scalar_manager,
                    )
                )

            self.replace_abstraction_top(_RowEntityPhiTensor(), rows=new_list)  # type: ignore

        elif ndept and entities is not None and len(entities) == self.shape[0]:
            class_type = _SingleEntityPhiTensor()
            entity_list = EntityList(one_hot_lookup, entities_indexed)

            if isinstance(min_val, (bool, int, float)):
                min_vals = np.array(min_val).ravel()  # make it 1D
            else:
                raise Exception(
                    "min_val should be either float,int,bool got "
                    + str(type(min_val))
                    + " instead."
                )

            if isinstance(max_val, (bool, int, float)):
                max_vals = np.array(max_val).ravel()  # make it 1D
            else:
                raise Exception(
                    "min_val should be either float,int,bool got "
                    + str(type(max_val))
                    + " instead."
                )

            if min_vals.shape != self.child.shape:
                min_vals = lazyrepeatarray(min_vals, self.child.shape)

            if max_vals.shape != self.child.shape:
                max_vals = lazyrepeatarray(max_vals, self.child.shape)

            self.replace_abstraction_top(
                tensor_type=_NDimEntityPhiTensor(),
                child=self.child,
                min_vals=min_vals,
                max_vals=max_vals,
                entities=entity_list,  # type: ignore
                row_type=class_type,  # type: ignore
            )  # type: ignore

        # TODO: if there's element-level entities - push all elements with PhiScalars
        else:
            raise Exception(
                "If you're passing in mulitple entities, please pass in one entity per row."
            )

        return self



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/python/int.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1aproto/lib/python/int.proto\x12\x0fsyft.lib.python\x1a%proto/core/common/common_object.proto"6\n\x03Int\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x03\x12!\n\x02id\x18\x02 \x01(\x0b\x32\x15.syft.core.common.UIDb\x06proto3'
)


_INT = DESCRIPTOR.message_types_by_name["Int"]
Int = _reflection.GeneratedProtocolMessageType(
    "Int",
    (_message.Message,),
    {
        "DESCRIPTOR": _INT,
        "__module__": "proto.lib.python.int_pb2"
        # @@protoc_insertion_point(class_scope:syft.lib.python.Int)
    },
)
_sym_db.RegisterMessage(Int)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _INT._serialized_start = 86
    _INT._serialized_end = 140
# @@protoc_insertion_point(module_scope)



# future
from __future__ import annotations

# stdlib
import sys
from typing import Any
from typing import Dict
from typing import List
from typing import Union

# third party
import pandas as pd
import requests

# relative
from . import login
from .core.node.common.client import Client
from .grid import GridURL
from .logger import error
from .logger import warning

NETWORK_REGISTRY_URL = (
    "https://raw.githubusercontent.com/OpenMined/NetworkRegistry/main/networks.json"
)
NETWORK_REGISTRY_REPO = "https://github.com/OpenMined/NetworkRegistry"


class NetworkRegistry:
    def __init__(self) -> None:
        self.networks: List[Dict] = []
        try:
            response = requests.get(NETWORK_REGISTRY_URL)
            network_json = response.json()
            self.all_networks = network_json["networks"]
        except Exception as e:
            warning(
                f"Failed to get Network Registry, go checkout: {NETWORK_REGISTRY_REPO}. {e}"
            )

    @property
    def online_networks(self) -> List[Dict]:
        online_networks = list()

        an = self.all_networks

        for i, network in enumerate(an):
            sys.stdout.write(
                "\rChecking network availability: " + str(i + 1) + " of " + str(len(an))
            )
            url = "http://" + network["host_or_ip"] + ":" + str(network["port"]) + "/"
            try:
                res = requests.get(url, timeout=0.5)
                online = "This is a PyGrid Network node." in res.text
            except Exception:
                online = False

            if online:
                online_networks.append(network)
        sys.stdout.write("\r                                             ")
        return online_networks

    def _repr_html_(self) -> str:
        on = self.online_networks
        if len(on) == 0:
            return "(no networks online - try syft.networks.all_networks to see offline networks)"
        return pd.DataFrame(on)._repr_html_()

    def create_client(self, network: Dict[str, Any]) -> Client:
        try:
            port = int(network["port"])
            protocol = network["protocol"]
            host_or_ip = network["host_or_ip"]
            grid_url = GridURL(port=port, protocol=protocol, host_or_ip=host_or_ip)
            return login(url=str(grid_url), port=port)
        except Exception as e:
            error(f"Failed to login with: {network}. {e}")
            raise e

    def __getitem__(self, key: Union[str, int]) -> Client:
        if isinstance(key, int):
            return self.create_client(network=self.online_networks[key])
        else:
            on = self.online_networks
            for network in on:
                if network["name"] == key:
                    return self.create_client(network=network)
        raise KeyError(f"Invalid key: {key} for {on}")



# stdlib
from typing import Any
from typing import List

# third party
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Query

# relative
from ..exceptions import EnvironmentNotFoundError
from ..node_table.environment import Environment
from ..node_table.user_environment import UserEnvironment
from .database_manager import DatabaseManager


class EnvironmentManager(DatabaseManager):

    schema = Environment
    user_env_association_schema = UserEnvironment

    def __init__(self, database: Engine) -> None:
        super().__init__(schema=EnvironmentManager.schema, db=database)
        self._association_schema = EnvironmentManager.user_env_association_schema

    def association(self, user_id: str, env_id: str) -> None:
        new_association_obj = self._association_schema(user=user_id, environment=env_id)
        self.db.session.add(new_association_obj)
        self.db.session.commit()

    def get_environments(self, **kwargs: Any) -> List[Environment]:
        objects = (
            self.db.session.query(self._association_schema).filter_by(**kwargs).all()
        )
        return objects

    def get_all_associations(self) -> List[UserEnvironment]:
        return list(self.db.session.query(self._association_schema).all())

    def delete_associations(self, environment_id: int) -> None:
        # Delete User environment Association
        associations = (
            self.db.session.query(self._association_schema)
            .filter_by(environment=environment_id)
            .all()
        )
        for association in associations:
            self.db.session.delete(association)

        self.db.session.commit()

    def first(self, **kwargs: Any) -> Environment:
        result = super().first(**kwargs)
        if not result:
            raise EnvironmentNotFoundError
        return result

    def query(self, **kwargs: Any) -> Query:
        results = super().query(**kwargs)
        return results

    def set(self, id: int, **kwargs: Any) -> None:
        self.modify({"id": id}, {**kwargs})



"""The purpose of this service is to inform lower-level devices
of changes in the hierarchy above them. For example, if a Domain
registers within a new Network or if Device registers within
a new Domain, all the other child node will need to know this
information to populate complete addresses into their clients."""

# stdlib
from typing import List

# third party
from nacl.signing import VerifyKey

# relative
from ......logger import debug
from ......logger import traceback
from ....abstract.node import AbstractNode
from ..auth import service_auth
from ..node_service import ImmediateNodeServiceWithoutReply
from .heritage_update_messages import HeritageUpdateMessage

# TODO: change all message names in syft to have "WithReply" or "WithoutReply"
# at the end of the name


class HeritageUpdateService(ImmediateNodeServiceWithoutReply):
    @staticmethod
    @service_auth(root_only=True)
    def process(
        node: AbstractNode, msg: HeritageUpdateMessage, verify_key: VerifyKey
    ) -> None:
        debug(
            f"> Executing {HeritageUpdateService.pprint()} {msg.pprint} on {node.pprint}"
        )
        addr = msg.new_ancestry_address

        if addr.network is not None:
            node.network = addr.network
        if addr.domain is not None:
            node.domain = addr.domain
        if addr.device is not None:
            node.device = addr.device

        for node_client in node.known_child_nodes:
            try:
                # TODO: Client (and possibly Node) should subclass from StorableObject
                location_id = node_client.address.target_id.id
                msg.address = node_client.address
                try:
                    in_memory_client = node.in_memory_client_registry[location_id]
                    # we need to sign here with the current node not the destination side
                    in_memory_client.send_immediate_msg_without_reply(msg=msg)
                    debug(f"> Flowing {msg.pprint} to {addr.target_emoji()}")
                    return None
                except Exception as e:
                    debug(f"{location_id} not on nodes in_memory_client. {e}")
                    pass
            except Exception as e:
                traceback(e)

    @staticmethod
    def message_handler_types() -> List[type]:
        return [HeritageUpdateMessage]



# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Optional
from typing import Type

# third party
from nacl.signing import VerifyKey

# relative
from ...abstract.node_service_interface import NodeServiceInterface
from ..node_service.generic_payload.syft_message import NewSyftMessage as SyftMessage


class BinaryOperation:
    """Executes an operation between two operands."""

    def __init__(self, op1: Any, op2: Any, operator: Any) -> None:
        self.op1 = op1
        self.op2 = op2
        self.operator = operator

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        op1 = self.op1(*args, **kwargs)
        op2 = self.op2(*args, **kwargs)
        return self.operator(op1, op2)


class UnaryOperation:
    """Executes an operation on a single operand."""

    def __init__(self, op1: Any, operator: Any) -> None:
        self.op1 = op1
        self.operator = operator

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        op1 = self.op1(*args, **kwargs)
        return self.operator(op1)


class AND:
    """Implements the `AND` functionality on a set of permission classes."""

    def __init__(self, op1: Type[BasePermission], op2: Type[BasePermission]):
        self.op1 = op1
        self.op2 = op2

    def has_permission(self, *args: Any, **kwargs: Any) -> bool:
        return self.op1.has_permission(*args, **kwargs) and self.op2.has_permission(
            *args, **kwargs
        )


class OR:
    """Implements the `OR` functionality on a set of permission classes."""

    def __init__(self, op1: Type[BasePermission], op2: Type[BasePermission]):
        self.op1 = op1
        self.op2 = op2

    def has_permission(self, *args: Any, **kwargs: Any) -> bool:
        return self.op1.has_permission(*args, **kwargs) or self.op2.has_permission(
            *args, **kwargs
        )


class NOT:
    """Implements the `NOT` functionality on a permission class."""

    def __init__(self, op1: Type[BasePermission]):
        self.op1 = op1

    def has_permission(self, *args: Any, **kwargs: Any) -> bool:
        return not self.op1.has_permission(*args, **kwargs)


class BasePermissionMetaclass(type):
    """A metaclass to allow composition between different permission classes."""

    def __and__(self, other: Type[BasePermission]) -> BinaryOperation:
        return BinaryOperation(self, other, AND)

    def __or__(self, other: Type[BasePermission]) -> BinaryOperation:  # type: ignore
        return BinaryOperation(self, other, OR)

    def __rand__(self, other: Type[BasePermission]) -> BinaryOperation:
        return BinaryOperation(other, self, AND)

    def __ror__(self, other: Type[BasePermission]) -> BinaryOperation:  # type: ignore
        return BinaryOperation(other, self, OR)

    def __invert__(self) -> UnaryOperation:
        return UnaryOperation(self, NOT)


class BasePermission(metaclass=BasePermissionMetaclass):
    """A base class from which all permission classes should inherit."""

    def has_permission(
        self,
        msg: SyftMessage,
        node: NodeServiceInterface,
        verify_key: Optional[VerifyKey],
    ) -> bool:
        """Return `True` if permission is granted, `False` otherwise."""
        raise NotImplementedError



"""A batched garbage collection heuristics.

Deleting a remote object using this logic might not send a message indicating
to the Data Owner to remove the object from the store.

Instead, it would collect the id of that object to be deleted and would only
send the "delete message" when a specific threshold limit is reached. At this
point the Data Owner will loop through all the ids from the received message and
will remove each object from the store.
"""
# stdlib
from typing import List
from typing import Optional
from typing import TYPE_CHECKING

# third party
from typing_extensions import final

# relative
from ...common.uid import UID
from ...node.common.action.garbage_collect_batched_action import (
    GarbageCollectBatchedAction,
)
from ..pointer import Pointer
from .gc_strategy import GCStrategy

if TYPE_CHECKING:

    # relative
    from ...node.common.client import Client


@final
class GCBatched(GCStrategy):
    """The GCBatched Strategy."""

    __slots__ = [
        "obj_ids",
        "threshold",
        "client",
    ]

    client: Optional["Client"]
    obj_ids: List[UID]
    threshold: int

    def __init__(self, threshold: int = 10) -> None:
        """Construct the GCBatched Strategy.

        Args:
            threshold (int): the threshold after which a message
                would be sent to delete all the objects that were cached
        Return:
            None
        """
        self.obj_ids = []
        self.threshold = threshold
        self.client = None

    def reap(self, pointer: Pointer) -> None:
        """
        Check if we passed the threshold of objects that we should cache.
        If yes, then send a message to delete the objects.
        If no, cache the object id to be deleted when there are more ids
        collected.

        Args:
            pointer (Pointer): Pointer to the object that should get deleted

        Return:
            None
        """

        self.obj_ids.append(pointer.id_at_location)

        nr_objs_client = len(self.obj_ids)

        if nr_objs_client >= self.threshold:
            # Check the local threshold for the items that are kept for a client
            msg = GarbageCollectBatchedAction(
                ids_at_location=self.obj_ids, address=pointer.client.address
            )

            pointer.client.send_eventual_msg_without_reply(msg)
            self.obj_ids = []

        self.client = pointer.client

    def __del__(self) -> None:
        """Send a GarbageCollectBatchedAction to the client such that all the
        objects that are cached to be deleted would be deleted.
        """
        if self.client is None:
            return

        msg = GarbageCollectBatchedAction(
            ids_at_location=self.obj_ids, address=self.client.address
        )

        self.client.send_eventual_msg_without_reply(msg)



# stdlib
from typing import Any

# relative
from .fixed_precision_tensor import FixedPrecisionTensor
from .manager import TensorChainManager
from .passthrough import SupportedChainType  # type: ignore


class FixedPrecisionTensorAncestor(TensorChainManager):
    def fix_precision(self, base: int = 10, precision: int = 3) -> Any:
        self.child: SupportedChainType = FixedPrecisionTensor(
            base=base, precision=precision, value=self.child  # type: ignore
        )
        return self

    def decode(self) -> Any:
        if not isinstance(self.child, FixedPrecisionTensor):
            raise ValueError(f"self.child should be FPT but is {type(self.child)}")

        res = self.child.decode()
        return res



# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey

# syft absolute
import syft as sy

# relative
from ..... import lib
from ..... import logger
from .....logger import traceback_and_raise
from .....proto.core.node.common.action.run_class_method_smpc_pb2 import (
    RunClassMethodSMPCAction as RunClassMethodSMPCAction_PB,
)
from ....common.serde.serializable import serializable
from ....common.uid import UID
from ....io.address import Address
from ....store.storeable_object import StorableObject
from ...abstract.node import AbstractNode
from .common import ImmediateActionWithoutReply
from .greenlets_switch import retrieve_object

if TYPE_CHECKING:
    # relative
    from .smpc_action_message import SMPCActionMessage


@serializable()
class RunClassMethodSMPCAction(ImmediateActionWithoutReply):
    """
    When executing a RunClassMethodSMPCAction, a list of SMPCActionMessages is sent to the
    to the same node (using a rabbitMQ)

    Attributes:
         path: the dotted path to the method to call
         _self: a pointer to the object which the method should be applied to.
         args: args to pass to the function. They should be pointers to objects
            located on the :class:`Node` that will execute the action.
         kwargs: kwargs to pass to the function. They should be pointers to objects
            located on the :class:`Node` that will execute the action.
    """

    def __init__(
        self,
        path: str,
        _self: Any,
        args: List[Any],
        kwargs: Dict[Any, Any],
        id_at_location: UID,
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        self.path = path
        self._self = _self
        self.args = args
        self.kwargs = kwargs
        self.id_at_location = id_at_location
        # logging needs .path to exist before calling
        # this which is why i've put this super().__init__ down here
        super().__init__(address=address, msg_id=msg_id)

    @staticmethod
    def intersect_keys(
        left: Dict[VerifyKey, UID], right: Dict[VerifyKey, UID]
    ) -> Dict[VerifyKey, UID]:
        # get the intersection of the dict keys, the value is the request_id
        # if the request_id is different for some reason we still want to keep it,
        # so only intersect the keys and then copy those over from the main dict
        # into a new one
        intersection = set(left.keys()).intersection(right.keys())
        # left and right have the same keys
        return {k: left[k] for k in intersection}

    @property
    def pprint(self) -> str:
        return f"RunClassMethodSMPCAction({self.path})"

    def __repr__(self) -> str:
        method_name = self.path.split(".")[-1]
        self_name = self._self.__class__.__name__
        arg_names = ",".join([a.__class__.__name__ for a in self.args])
        kwargs_names = ",".join(
            [f"{k}={v.__class__.__name__}" for k, v in self.kwargs.items()]
        )
        return f"RunClassMethodSMPCAction {self_name}.{method_name}({arg_names}, {kwargs_names})"

    def execute_action(self, node: AbstractNode, verify_key: VerifyKey) -> None:
        # relative
        from . import smpc_action_functions
        from ..... import Tensor
        from .smpc_action_message import SMPCActionMessage
        from .smpc_action_seq_batch_message import SMPCActionSeqBatchMessage

        resolved_self = retrieve_object(node, self._self.id_at_location, self.path)

        result_read_permissions = resolved_self.read_permissions

        resolved_args = list()
        tag_args = []
        for arg in self.args:
            r_arg = retrieve_object(node, arg.id_at_location, self.path)

            # TODO: Think of a way to free the memory
            # del node.store[arg.id_at_location]
            result_read_permissions = self.intersect_keys(
                result_read_permissions, r_arg.read_permissions
            )
            resolved_args.append(r_arg.data)
            tag_args.append(r_arg)

        resolved_kwargs = {}
        tag_kwargs = {}
        for arg_name, arg in self.kwargs.items():
            r_arg = retrieve_object(node, arg.id_at_location, self.path)
            # TODO: Think of a way to free the memory
            # del node.store[arg.id_at_location]
            result_read_permissions = self.intersect_keys(
                result_read_permissions, r_arg.read_permissions
            )
            resolved_kwargs[arg_name] = r_arg.data
            tag_kwargs[arg_name] = r_arg

        (
            upcasted_args,
            upcasted_kwargs,
        ) = lib.python.util.upcast_args_and_kwargs(resolved_args, resolved_kwargs)

        method_name = self.path.split(".")[-1]
        value = resolved_self.data
        if isinstance(value, Tensor):
            nr_parties = value.child.child.nr_parties
            rank = value.child.child.rank
        else:
            nr_parties = value.nr_parties
            rank = value.rank

        seed_id_locations = resolved_kwargs.get("seed_id_locations", None)
        if seed_id_locations is None:
            raise ValueError(
                "Expected 'seed_id_locations' to be in the kwargs to generate id_at_location in a deterministic matter"
            )

        resolved_kwargs.pop("seed_id_locations")

        actions_generator = smpc_action_functions.get_action_generator_from_op(
            operation_str=method_name, nr_parties=nr_parties
        )
        args_id = [arg.id_at_location for arg in self.args]

        # TODO: For the moment we don't run any SMPC operation that provides any kwarg
        kwargs = {
            "seed_id_locations": int(seed_id_locations),
            "node": node,
        }

        # Get the list of actions to be run
        # TODO : Remove client as we do not use it now.
        actions: Union[List[SMPCActionMessage], SMPCActionSeqBatchMessage]
        actions = actions_generator(*args_id, **kwargs)  # type: ignore

        if isinstance(actions, (list, tuple)) and isinstance(
            actions[0], SMPCActionMessage
        ):
            actions = SMPCActionMessage.filter_actions_after_rank(rank, actions)
            for action in actions:
                RunClassMethodSMPCAction.execute_smpc_action(node, action, verify_key)
        elif isinstance(actions, SMPCActionSeqBatchMessage):
            msg = actions
            while msg.smpc_actions:
                action = msg.smpc_actions[0]
                RunClassMethodSMPCAction.execute_smpc_action(node, action, verify_key)
                del msg.smpc_actions[0]

    @staticmethod
    def execute_smpc_action(
        node: AbstractNode, msg: "SMPCActionMessage", verify_key: VerifyKey
    ) -> None:
        # relative
        from .smpc_action_functions import _MAP_ACTION_TO_FUNCTION

        func = _MAP_ACTION_TO_FUNCTION[msg.name_action]
        store_object_self = node.store.get(key=msg.self_id)
        if store_object_self is None:
            raise KeyError("Object not already in store")

        _self = store_object_self.data
        args = [node.store.get(arg_id).data for arg_id in msg.args_id]

        kwargs = {}  # type: ignore
        for key, kwarg_id in msg.kwargs_id.items():
            data = node.store.get(kwarg_id).data
            if data is None:
                raise KeyError(f"Key {key} is not available")

            kwargs[key] = data
        kwargs = {**kwargs, **msg.kwargs}
        (
            upcasted_args,
            upcasted_kwargs,
        ) = lib.python.util.upcast_args_and_kwargs(args, kwargs)
        logger.warning(func)

        if msg.name_action in {"spdz_multiply", "spdz_mask"}:
            result = func(_self, *upcasted_args, **upcasted_kwargs, node=node)
        elif msg.name_action == "local_decomposition":
            result = func(
                _self,
                *upcasted_args,
                **upcasted_kwargs,
                node=node,
                read_permissions=store_object_self.read_permissions,
            )
        else:
            result = func(_self, *upcasted_args, **upcasted_kwargs)

        if lib.python.primitive_factory.isprimitive(value=result):
            # Wrap in a SyPrimitive
            result = lib.python.primitive_factory.PrimitiveFactory.generate_primitive(
                value=result, id=msg.id_at_location
            )
        else:
            # TODO: overload all methods to incorporate this automatically
            if hasattr(result, "id"):
                try:
                    if hasattr(result, "_id"):
                        # set the underlying id
                        result._id = msg.id_at_location
                    else:
                        result.id = msg.id_at_location

                    if result.id != msg.id_at_location:
                        raise AttributeError("IDs don't match")
                except AttributeError as e:
                    err = f"Unable to set id on result {type(result)}. {e}"
                    traceback_and_raise(Exception(err))

        if not isinstance(result, StorableObject):
            result = StorableObject(
                id=msg.id_at_location,
                data=result,
                read_permissions=store_object_self.read_permissions,
            )

        node.store[msg.id_at_location] = result

    def _object2proto(self) -> RunClassMethodSMPCAction_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: RunClassMethodSMPCAction_PB

        .. note::
            This method is purely an internal method. Please use sy.serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """

        return RunClassMethodSMPCAction_PB(
            path=self.path,
            _self=sy.serialize(self._self),
            args=list(map(lambda x: sy.serialize(x), self.args)),
            kwargs={k: sy.serialize(v) for k, v in self.kwargs.items()},
            id_at_location=sy.serialize(self.id_at_location),
            address=sy.serialize(self.address),
            msg_id=sy.serialize(self.id),
        )

    @staticmethod
    def _proto2object(proto: RunClassMethodSMPCAction_PB) -> "RunClassMethodSMPCAction":
        """Creates a ObjectWithID from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of RunClassMethodSMPCAction
        :rtype: RunClassMethodSMPCAction

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return RunClassMethodSMPCAction(
            path=proto.path,
            _self=sy.deserialize(blob=proto._self),
            args=list(map(lambda x: sy.deserialize(blob=x), proto.args)),
            kwargs={k: sy.deserialize(blob=v) for k, v in proto.kwargs.items()},
            id_at_location=sy.deserialize(blob=proto.id_at_location),
            address=sy.deserialize(blob=proto.address),
            msg_id=sy.deserialize(blob=proto.msg_id),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return RunClassMethodSMPCAction_PB
